#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

__name__ = 'woltka'
__description__ = 'versatile metagenome classifier'
__version__ = '0.1.4'
__license__ = 'BSD-3-Clause'
__author__ = 'Qiyun Zhu'
__email__ = 'qiyunzhu@gmail.com'
__url__ = 'https://github.com/qiyunzhu/woltka'



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

"""Functions for matching reads and genes using an ordinal system.
"""

from collections import defaultdict
from itertools import chain
from bisect import bisect

from .align import infer_align_format, assign_parser


def ordinal_mapper(fh, coords, idmap, fmt=None, n=1000000, th=0.8,
                   prefix=False):
    """Read an alignment file and match reads and genes in an ordinal system.

    Parameters
    ----------
    fh : file handle
        Alignment file to parse.
    coords : dict of list
        Gene coordinates table.
    idmap : dict of list
        Gene identifiers.
    fmt : str, optional
        Alignment file format.
    n : int, optional
        Number of lines per chunk.
    th : float
        Minimum threshold of overlap length : alignment length for a match.
    prefix : bool
        Prefix gene IDs with nucleotide IDs.

    See Also
    --------
    align.plain_mapper

    Yields
    ------
    tuple of str
        Query queue.
    dict of set of str
        Subject(s) queue.
    """
    # determine file format
    fmt, head = (fmt, []) if fmt else infer_align_format(fh)

    # assign parser for given format
    parser = assign_parser(fmt, ext=True)

    # cached list of query Ids for reverse look-up
    # gene Ids are unique, but read Ids can have duplicates (i.e., one read is
    # mapped to multiple loci on a genome), therefore an incremental integer
    # here replaces the original read Id as its identifer
    rids = []
    rid_append = rids.append

    # cached map of read to coordinates
    locmap = defaultdict(list)

    def flush():
        """Match reads in current chunk with genes from all nucleotides.

        Returns
        -------
        tuple of str
            Query queue.
        dict of set of str
            Subject(s) queue.
        """
        # master read-to-gene(s) map
        res = defaultdict(set)

        # iterate over nucleotides
        for nucl, locs in locmap.items():

            # it's possible that no gene was annotated on the nucleotide
            try:
                glocs = coords[nucl]
            except KeyError:
                continue

            # get reference to gene identifiers
            gids = idmap[nucl]

            # append prefix if needed
            pfx = nucl + '_' if prefix else ''

            # execute ordinal algorithm when reads are many
            # 8 (5+ reads) is an empirically determined cutoff
            if len(locs) > 8:

                # merge and sort coordinates
                # question is to add unsorted read coordinates into pre-sorted
                # gene coordinates
                # Python's Timsort algorithm is efficient for this task
                queue = sorted(chain(glocs, locs))

                # map reads to genes using the core algorithm
                for read, gene in match_read_gene(queue):

                    # add read-gene pairs to the master map
                    res[rids[read]].add(pfx + gids[gene])

            # execute naive algorithm when reads are few
            else:
                for read, gene in match_read_gene_quart(glocs, locs):
                    res[rids[read]].add(pfx + gids[gene])

        # return matching read Ids and gene Ids
        return res.keys(), res.values()

    this = None  # current query Id
    target = n   # target line number at end of current chunk

    # parse alignment file
    for i, row in enumerate(parser(chain(iter(head), fh))):
        query, subject, _, length, beg, end = row[:6]

        # skip if length is not available or zero
        if not length:
            continue

        # when query Id changes and chunk limits has been reached
        if query != this and i >= target:

            # flush: match currently cached reads with genes and yield
            yield flush()

            # re-initiate read Ids, length map and location map
            rids = []
            rid_append = rids.append
            locmap = defaultdict(list)

            # next target line number
            target = i + n

        # append read Id, alignment length and location
        idx = len(rids)
        rid_append(query)

        # effective length = length * th
        # -int(-x // 1) is equivalent to math.ceil(x) but faster
        # this value must be >= 1
        locmap[subject].extend((
            (beg << 48) + (-int(-length * th // 1) << 31) + idx,
            (end << 48) + idx))
        this = query

    # final flush
    yield flush()


def ordinal_parser_dummy(fh, parser):
    """Alignment parsing functionalities stripped from for `ordinal_mapper`.

    Parameters
    ----------
    fh : file handle
        Alignment file to parse.
    parser : callable
        Function to parse alignment lines of certain format.

    Returns
    -------
    list of str
        Read Ids in same order as in alignment file.
    defaultdict of dict of int
        Map of read indices to alignment lengths per nucleotide.
    defaultdict of list of (int, bool, bool, str)
        Flattened list of read coordinates per nucleotide.

    See Also
    --------
    ordinal_mapper
    match_read_gene_dummy
    .tests.test_ordinal.OrdinalTests.test_ordinal_parser_dummy

    Notes
    -----
    This is a dummy function only for test and demonstration purpose but not
    called anywhere in the program. See its unit test for details.
    """
    rids = []
    lenmap = defaultdict(dict)
    locmap = defaultdict(list)

    for row in parser(fh):
        query, subject, _, length, start, end = row[:6]
        idx = len(rids)
        rids.append(query)
        lenmap[subject][idx] = length
        locmap[subject].extend((
            (start, True, False, idx),
            (end,  False, False, idx)))

    return rids, lenmap, locmap


def load_gene_coords(fh, sort=False):
    """Read coordinates of genes on genomes.

    Parameters
    ----------
    fh : file handle
        Gene coordinates file.
    sort : bool, optional
        Whether sort gene coordinates.

    Returns
    -------
    dict of int
        Binarized gene coordinate information per nucleotide.
    dict of list of str
        Gene IDs.
    bool
        Whether there are duplicate gene IDs.

    See Also
    --------
    match_read_gene

    Notes
    -----
    This data structure is central to this algorithm. Starting and ending
    coordinates of each gene are separated and flattened into a sorted list.
    which enables only one round of list traversal for the entire set of genes
    plus reads.

    See the docstring of `match_read_gene` for details.
    """
    coords = {}
    queue_extend = None

    idmap = {}
    gids = None
    gids_append = None

    isdup = None
    used = set()
    used_add = used.add

    for line in fh:

        # ">" or "#" indicates genome (nucleotide) name
        c0 = line[0]
        if c0 in '>#':

            # double ">" or "#" indicates genome name, which serves as
            # a super group of subsequent nucleotide names; to be ignored
            if line[1] != c0:
                nucl = line[1:].strip()
                coords[nucl] = []
                queue_extend = coords[nucl].extend
                gids = idmap[nucl] = []
                gids_append = gids.append
        else:
            x = line.rstrip().split('\t')

            # begin and end positions are based on genome (nucleotide)
            try:
                beg, end = int(x[1]), int(x[2])
            except (IndexError, ValueError):
                raise ValueError(
                    f'Cannot extract coordinates from line: "{line}".')
            idx = len(gids)
            gene = x[0]
            gids_append(gene)
            if beg > end:
                beg, end = end, beg
            queue_extend(((beg << 48) + (3 << 30) + idx,
                          (end << 48) + (1 << 30) + idx))

            # check duplicate
            if isdup is None:
                if gene in used:
                    isdup = True
                else:
                    used_add(gene)

    # sort gene coordinates per nucleotide
    if sort:
        for queue in coords.values():
            queue.sort()

    return coords, idmap, isdup or False


def match_read_gene(queue):
    """Associate reads with genes based on a sorted queue of coordinates.

    Parameters
    ----------
    queue : list of int
        Sorted queue of coordinates.

    Yields
    ------
    int
        Read index.
    int
        Gene index.

    See Also
    --------
    load_gene_coords
    match_read_gene_dummy
    .tests.test_ordinal.OrdinalTests.test_match_read_gene_dummy

    Notes
    -----
    This algorithm is the core of this module. It uses a flattened, sorted
    list to store starting and ending coordinates of both genes and reads.
    Only one round of traversal (O(n)) of this list is needed to accurately
    find all gene-read matches.

    Refer to its unit test `test_match_read_gene` for an illustrated example.

    This function is the most compute-intensive step in the entire analysis,
    therefore it has been deeply optimized to increase performance wherever
    possible. Notably, it extensively uses bitwise operations to extract
    multiple pieces of information from a single integer.

    Specifically, each coordinate (an integer) has the following information
    (from right to left):

    - Bits  1-30: Index of gene / read (30 bits, max.: 1,073,741,823).
    - Bits    31: Whether it is a gene (1) or a read (0) (1 bit).
    - Bits 32-58: Whether it is the start (positive) or end (0) of a gene /
                  read. If start, the value represents the effective length of
                  an alignment if it's a read, or 1 if it's a gene (17 bits,
                  max.: 131,071).
    - Bits 59-  : Coordinate (position on the genome, nt) (unlimited)

    The Python code to extract these pieces of information is:

    - Coordinate:       `code >> 48`
    - Effective length: `code >> 31 & (1 << 17) - 1`, or `code >> 31 & 131071`
    - Gene or read:     `code & (1 << 30)`
    - Gene/read index:  `code & (1 << 30) - 1`

    Note: Repeated bitwise operations are usually more efficient that a single
    bitwise operation assigned to a new variable.
    """
    genes = {}  # current genes cache
    reads = {}  # current reads cache

    # cache method references
    genes_items = genes.items
    reads_items = reads.items

    genes_pop = genes.pop
    reads_pop = reads.pop

    # walk through flattened queue of reads and genes
    for code in queue:

        # if this is a gene,
        if code & (1 << 30):

            # when a gene begins,
            # if code >> 31 & 131071:
            if code & (1 << 31):

                # add its index and coordinate to cache
                genes[code & (1 << 30) - 1] = code >> 48

            # when a gene ends,
            else:

                # find gene start
                gloc = genes_pop(code & (1 << 30) - 1)

                # check cached reads for matches
                for rid, rloc in reads_items():

                    # is a match if read/gene overlap is long enough
                    #   code >> 48:     read end coordinate
                    #   gloc:           gene start coordinate
                    #   rloc >> 17`:    read start coordinate
                    #   rloc & 131071`: effective length - 1
                    if (code >> 48) - max(gloc, rloc >> 17) >= rloc & 131071:
                        yield rid, code & (1 << 30) - 1

        # if this is a read,
        else:

            # when a read begins,
            if code >> 31 & 131071:

                # add its index, coordinate and effective length - 1 to cache
                # the latter two are stored as a single integer
                reads[code & (1 << 30) - 1] = (code >> 31) - 1

            # when a read ends,
            else:

                # find read start and effective length
                rloc = reads_pop(code & (1 << 30) - 1)

                # check cached genes
                # a potential optimization is to pre-calculate `rloc >> 17`
                #   and `(code >> 48) - (rloc & 131071)`, however, it is not
                #   worth the overhead in real applications because there is
                #   usually zero or one gene in cache
                # another potential optimization is to replace `max` (which is
                #   a function call with a ternary operator, but one needs the
                #   first optimization prior to this
                for gid, gloc in genes_items():

                    # same as above
                    if (code >> 48) - max(gloc, rloc >> 17) >= rloc & 131071:
                        yield code & (1 << 30) - 1, gid


def match_read_gene_naive(geneque, readque):
    """Associate reads with genes using a naive approach, which performs
    nested iteration over genes and reads.

    Parameters
    ----------
    geneque : list of int
        Sorted queue of genes.
    readque : list of int
        Paired queue of reads.

    Yields
    ------
    int
        Read index.
    int
        Gene index.

    See Also
    --------
    match_read_gene

    Notes
    -----
    This is a reference implementation of the naive nested iteration method
    with limited optimization. It is O(nm), where n and m are the numbers of
    genes and reads, respectively. It should be much slower than the ordinal
    method. However, when the number of reads is small, it may be efficient
    because it saves the sorting step.
    """
    # only genes are cached
    genes = {}
    genes_pop = genes.pop

    # iterate over reads (paired)
    it = iter(readque)
    for x, y in zip(it, it):

        # pre-calculate read metrics
        rid = x & (1 << 30) - 1     # index
        beg = x >> 48               # start coordinate
        end = y >> 48               # end coordinate
        L = (x >> 31) - 1 & 131071  # effective length - 1

        # iterate over genes (ordinal)
        for code in geneque:

            # add gene to cache
            if code & (1 << 31):
                genes[code & (1 << 30) - 1] = code >> 48

            # check overlap while removing gene from cache:
            # min(gene end, read end) - max(gene start, read start) >=
            # effective length - 1
            elif (min(code >> 48, end) -
                  max(genes_pop(code & (1 << 30) - 1), beg)) >= L:
                yield rid, code & (1 << 30) - 1


def match_read_gene_quart(geneque, readque):
    """Associate reads with genes by iterating between read region and nearer
    end of gene queue.

    Parameters
    ----------
    geneque : list of int
        Sorted queue of genes.
    readque : list of int
        Paired queue of reads.

    Yields
    ------
    int
        Read index.
    int
        Gene index.

    See Also
    --------
    match_read_gene_naive

    Notes
    -----
    This optimized method is similar to the naive method, but it reduces the
    search space from n to n / 4, thus it significantly improves efficiency.

    The idea is that one only needs to iterate between the position where read
    can fit it, and the nearer end of the gene queue.

    Further reduction of search space is not possible, because genes may span
    over half of the genome, and they cannot be detected without iterating to
    end of the gene queue.

    This method uses bisection to find the insertion point of read start in the
    pre-sorted gene queue, which has O(logn) time.
    """
    n = len(geneque)  # entire search space
    mid = n // 2      # mid point

    # iterate over paired read starts and ends
    it = iter(readque)
    for x, y in zip(it, it):

        # pre-calculate read metrics
        rid = x & (1 << 30) - 1     # index
        beg = x >> 48               # start coordinate
        end = y >> 48               # end coordinate
        L = (x >> 31) - 1 & 131071  # effective length - 1

        # genes starting within read region
        # will record their coordinates
        within = {}
        within_pop = within.pop

        # read starts in left half of gene queue
        if x < geneque[mid]:

            # genes starting before read region
            # no need to record coordinates as overlap is not possible
            before = set()
            before_add = before.add
            before_remove = before.remove

            # locate read start using bisection
            # Python's `bisect` is more efficient than homebrew code
            i = bisect(geneque, x, hi=mid)

            # iterate over gene positions before read region
            for code in geneque[:i]:

                # add gene to cache when it starts
                if code & (1 << 31):
                    before_add(code & (1 << 30) - 1)

                # drop gene from cache when it ends
                else:
                    before_remove(code & (1 << 30) - 1)

            # iterate over gene positions after read start
            for code in geneque[i:]:

                # stop when exceeding read end
                if code > y:
                    break

                # when gene starts, add it and its coordinate to cache
                elif code & (1 << 31):
                    within[code & (1 << 30) - 1] = code >> 48

                # when gene ends, check overlap and remove it from cache
                else:
                    gid = code & (1 << 30) - 1

                    # most genes should start before read region
                    try:
                        before_remove(gid)

                    # if gene started within read region,
                    # overlap = gene end - gene start
                    except KeyError:
                        if (code >> 48) - within_pop(gid) >= L:
                            yield rid, gid

                    # if gene started before read region,
                    # overlap = gene end - read start
                    else:
                        if (code >> 48) - beg >= L:
                            yield rid, gid

            # yield all genes that started before read but not yet ended
            for gid in before:
                yield rid, gid

            # yield genes that started after read but not yet ended,
            # until overlap = read end - gene start is too short
            for gid, gloc in within.items():
                if end - gloc < L:
                    break
                yield rid, gid

        # read starts in right half of gene queue
        else:

            # genes starting after read region
            # no need to record coordinates as overlap is not possible
            after = set()
            after_add = after.add
            after_remove = after.remove

            # locate read start using bisection
            i = bisect(geneque, x, lo=mid + 1, hi=n)

            # iterate over gene positions within read region
            # for i in range(bisect(geneque, x), n):
            while i < n:
                code = geneque[i]

                # stop when exceeding read end
                if code > y:
                    break

                # when gene starts, add it and its coordinate to cache
                if code & (1 << 31):
                    within[code & (1 << 30) - 1] = code >> 48

                # when gene ends, check overlap and remove it from cache
                # gene end must be <= read end
                # if gene not found in cache (meaning gene started before read
                # region), use read start, otherwise use gene start
                elif (code >> 48) - within_pop(code & (1 << 30) - 1, beg) >= L:
                    yield rid, code & (1 << 30) - 1

                # move to next position
                i += 1

            # iterate over gene positions after read region
            for code in geneque[i:]:

                # add gene to cache when it starts
                if code & (1 << 31):
                    after_add(code & (1 << 30) - 1)

                # when gene ends,
                else:

                    # most remaining genes should start after read region
                    try:
                        after_remove(code & (1 << 30) - 1)

                    # otherwise, the gene spans over entire read region
                    except KeyError:

                        # check overlap while removing gene from cache
                        # gene end must be >= read end
                        # if gene not found in cache, use read start, otherwise
                        # use gene start
                        if end - within_pop(code & (1 << 30) - 1, beg) >= L:
                            yield rid, code & (1 << 30) - 1


def match_read_gene_dummy(queue, lens, th):
    """Associate reads with genes based on a sorted queue of coordinates.
    Parameters
    ----------
    queue : list of tuple of (int, bool, bool, int)
        Sorted queue of coordinates (location, start or end, gene or read,
        index).
    lens : dict
        Read-to-alignment length map.
    th : float
        Threshold for read/gene overlapping fraction.

    Yields
    ------
    int
        Read index.
    int
        Gene index.

    See Also
    --------
    match_read_gene
    .tests.test_ordinal.OrdinalTests.test_match_read_gene_dummy

    Notes
    -----
    This is a dummy function which is only for test and demonstration purpose,
    but is not called anywhere in the program.

    The formal function `match_read_gene` extensively uses bitwise operations
    and thus is hard to read. Therefore the current function, which represents
    the original form of prior to optimization, is retained.
    """
    genes = {}  # current genes
    reads = {}  # current reads

    # walk through flattened queue of reads and genes
    for loc, is_start, is_gene, idx in queue:
        if is_gene:

            # when a gene starts, added to gene cache
            if is_start:
                genes[idx] = loc

            # when a gene ends,
            else:

                # find gene start and remove it from cache
                gloc = genes.pop(idx)

                # check cached reads for matches
                for rid, rloc in reads.items():

                    # is a match if read/gene overlap is long enough
                    if loc - max(gloc, rloc) + 1 >= lens[rid] * th:
                        yield rid, idx

        # the same for reads
        else:
            if is_start:
                reads[idx] = loc
            else:
                rloc = reads.pop(idx)
                for gid, gloc in genes.items():
                    if loc - max(rloc, gloc) + 1 >= lens[idx] * th:
                        yield idx, gid


def calc_gene_lens(mapper):
    """Calculate gene lengths by start and end coordinates.

    Parameters
    ----------
    mapper : callable
        Ordinal mapper.

    Returns
    -------
    dict of int
        Mapping of genes to lengths.
    """
    res = {}
    prefix = mapper.keywords['prefix']
    idmap = mapper.keywords['idmap']
    for nucl, queue in mapper.keywords['coords'].items():
        idmap_ = idmap[nucl]
        nucl += '_'
        for code in queue:
            gid = idmap_[code & (1 << 30) - 1]
            if prefix:
                gid = nucl + gid
            if code >> 31 & 131071:
                res[gid] = 1 - (code >> 48)
            else:
                res[gid] += code >> 48
    return res



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

"""Features under the `tools` command group.
"""

from sys import exit
from os import listdir
from os.path import isdir, join, basename
import click

from .table import (
    read_table, table_shape, table_max_f, frac_table, divide_table,
    scale_table, round_table, write_table, filter_table, merge_tables,
    add_metacol, collapse_table, calc_coverage)
from .file import readzip, read_map_1st, read_map_many, read_map_all
from .util import scale_factor
from .tree import read_names


def normalize_wf(input_fp:  str,
                 output_fp: str,
                 sizes_fp:  str = None,
                 scale:     str = None,
                 digits:    int = None):
    """Workflow for normalizing a profile.

    Raises
    ------
    SystemExit
        Scale factor is not valid.
    SystemExit
        Feature not found in size map.
    """
    # read input profile
    table, fmt = read_table(input_fp)

    # estimate maximum decimal precision
    if digits is None:
        digits = table_max_f(table)

    # normalize to fractions
    if not sizes_fp:
        click.echo('Normalizing profile to fractions...', nl=False)
        table = frac_table(table)
        click.echo(' Done.')

    # filter profile by threshold
    else:
        click.echo(f'Reading feature sizes from file: {basename(sizes_fp)}...',
                   nl=False)
        with readzip(sizes_fp, {}) as f:
            sizes = {k: float(v) for k, v in read_map_1st(f)}
        click.echo(' Done.')
        click.echo('Normalizing profile by feature size...', nl=False)
        try:
            divide_table(table, sizes)
        except KeyError:
            exit('One or more features are not found in the size map.')
        click.echo(' Done.')

    # scale table values
    if scale:
        try:
            scale = scale_factor(scale)
        except ValueError:
            exit(f'"{scale}" is not a valid scale factor.')
        click.echo(f'Scaling profile by {scale} times...', nl=False)
        scale_table(table, scale)
        click.echo(' Done.')

    # round table values
    round_table(table, digits or None)

    # write normalized profile
    write_table(table, output_fp)
    click.echo('Normalized profile written.')


def filter_wf(input_fp:      str,
              output_fp:     str,
              min_count:     int = None,
              min_percent: float = None):
    """Workflow for filtering a profile based on a per-sample threshold.

    Raises
    ------
    SystemExit
        Neither threshold is provided.
    SystemExit
        Both thresholds are provided.
    SystemExit
        Percentage threshold >= 100.

    See Also
    --------
    .cli.filter_cmd
        Command-line arguments and help information.
    """
    # validate parameters
    if not any((min_count, min_percent)):
        exit('Please specify either minimum count or minimum percentage '
             'threshold.')
    if all((min_count, min_percent)):
        exit('Only one of minimum count or minimum percentage thresholds '
             'can be specified.')
    if min_percent and min_percent >= 100:
        exit('Minimum percentage threshold must be below 100.')

    # determine threshold
    th = min_count or min_percent / 100

    # read input profile
    table, fmt = read_table(input_fp)
    n = table_shape(table)[0]
    click.echo(f'Number of features before filtering: {n}.')

    # filter profile by threshold
    click.echo('Filtered profile...', nl=False)
    table = filter_table(table, th)
    click.echo(' Done.')
    n = table_shape(table)[0]
    click.echo(f'Number of features after filtering: {n}.')

    # write filtered profile
    write_table(table, output_fp)
    click.echo('Filtered profile written.')


def merge_wf(input_fps: list,
             output_fp:  str):
    """Workflow for merging multiple profiles into one.

    Raises
    ------
    SystemExit
        Only one profile is provided.
    SystemExit
        Found invalid profile.

    See Also
    --------
    .cli.merge_cmd
        Command-line arguments and help information.
    """
    click.echo('Reading profiles...')
    tables = []

    def _read_profile(fp):
        try:
            table = read_table(fp)[0]
        except ValueError:
            exit(f'Cannot parse {basename(fp)} as a profile.')
        n, m = table_shape(table)
        click.echo(f'  Read {basename(fp)}. Samples: {m}, features: {n}.')
        tables.append(table)

    for fp in input_fps:
        if isdir(fp):
            for fname in listdir(fp):
                _read_profile(join(fp, fname))
        else:
            _read_profile(fp)

    if len(tables) == 1:
        exit('Please provide two or more profiles.')
    click.echo(f'Done. Number of profiles read: {len(tables)}.')

    # estimate maximum decimal precision
    digits = max(map(table_max_f, tables))

    # merge profiles
    click.echo('Merging profiles...', nl=False)
    table = merge_tables(tables)
    click.echo(' Done.')
    n, m = table_shape(table)
    click.echo(f'Number of samples after merging: {m}.')
    click.echo(f'Number of features after merging: {n}.')

    # round table values
    round_table(table, digits or None)

    # write merged profile
    write_table(table, output_fp)
    click.echo('Merged profile written.')


def collapse_wf(input_fp:  str,
                map_fp:    str,
                output_fp: str,
                divide:   bool = False,
                field:     int = None,
                names_fp:  str = None):
    """Workflow for collapsing a profile based on many-to-many mapping.

    Raises
    ------
    SystemExit
        No relationship is found in mapping file.

    See Also
    --------
    .cli.collapse_cmd
        Command-line arguments and help information.
    """
    # read input profile
    table, fmt = read_table(input_fp)
    n = table_shape(table)[0]
    click.echo(f'Number of features before collapsing: {n}.')

    # maximum decimal precision
    digits = table_max_f(table)

    # read mapping file (many-to-many okay)
    with readzip(map_fp, {}) as f:
        mapping = read_map_many(f)
    if not mapping:
        exit(f'No source-target relationship is found in {basename(map_fp)}.')

    # convert field index
    if field:
        field -= 1

    # collapse profile by mapping
    click.echo('Collapsing profile...', nl=False)
    table = collapse_table(table, mapping, divide, field)
    click.echo(' Done.')
    n = table_shape(table)[0]
    click.echo(f'Number of features after collapsing: {n}.')

    # append feature names (optional)
    if names_fp:
        with readzip(names_fp, {}) as f:
            namedic = read_names(f)
        add_metacol(table, namedic, 'Name')

    # round table values
    round_table(table, digits or None)

    # write collapsed profile
    write_table(table, output_fp)
    click.echo('Collapsed profile written.')


def coverage_wf(input_fp:   str,
                map_fp:     str,
                output_fp:  str,
                threshold:  int = None,
                count:     bool = False,
                names_fp:   str = None):
    """Calculate coverage of table over feature groups given membership
    information.
    """
    # read input profile
    table, fmt = read_table(input_fp)
    n = table_shape(table)[0]
    click.echo(f'Number of features in profile: {n}.')

    # read mapping file (many-to-many okay)
    with readzip(map_fp, {}) as f:
        mapping = dict(read_map_all(f))
    if not mapping:
        exit(f'No group membership is found in {basename(map_fp)}.')

    # calculate group coverage
    click.echo('Calculating coverage...', nl=False)
    table = calc_coverage(table, mapping, threshold, count)
    click.echo(' Done.')
    n = table_shape(table)[0]
    click.echo(f'Number of feature groups with coverage: {n}.')

    # append feature group names (optional)
    if names_fp:
        with readzip(names_fp, {}) as f:
            namedic = read_names(f)
        add_metacol(table, namedic, 'Name')

    # write coverage table
    write_table(table, output_fp)
    click.echo('Coverage table written.')



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from os.path import join
from .plugin_setup import plugin
from ._format import (SeqAlnMapDirFmt,
                      BLAST6OutDirFmt,
                      SimpleMapDirFmt,
                      NCBINodesDirFmt,
                      GeneCoordDirFmt)


@plugin.register_transformer
def _1(ff: SeqAlnMapDirFmt) -> str:
    return join(str(ff.path), 'alignment.sam')


@plugin.register_transformer
def _2(ff: BLAST6OutDirFmt) -> str:
    return join(str(ff.path), 'alignment.b6o')


@plugin.register_transformer
def _3(ff: SimpleMapDirFmt) -> str:
    return join(str(ff.path), 'mapping.txt')


@plugin.register_transformer
def _4(ff: NCBINodesDirFmt) -> str:
    return join(str(ff.path), 'nodes.dmp')


@plugin.register_transformer
def _5(ff: GeneCoordDirFmt) -> str:
    return join(str(ff.path), 'coordinates.txt')



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

"""Standalone command-line interface (CLI) of the program.
"""

import click

from . import __version__
from .workflow import workflow
from .tools import normalize_wf, filter_wf, merge_wf, collapse_wf, coverage_wf


class NaturalOrderGroup(click.Group):
    """Natural ordering of click command groups.
    """
    def list_commands(self, ctx):
        return self.commands.keys()


@click.version_option(__version__)
@click.group(cls=NaturalOrderGroup)
def cli():
    pass  # pragma: no cover


# `classify` invokes the main classification workflow

@cli.command('classify')
# input and output
@click.option(
    '--input', '-i', 'input_fp', required=True, type=click.Path(
        exists=True, file_okay=True, dir_okay=True, allow_dash=True),
    help=('Path to input alignment file or directory of alignment files.'
          ' Enter "-" for stdin.'))
@click.option(
    '--output', '-o', 'output_fp', required=True,
    type=click.Path(writable=True),
    help='Path to output profile file or directory of profile files.')
# input files
@click.option(
    '--format', '-f', 'input_fmt',
    type=click.Choice(['b6o', 'sam', 'map'], case_sensitive=False),
    help=('Format of read alignments: "b6o": BLAST tabular format, "sam": SAM '
          'format, "map": simple map of query <tab> subject. If not specified,'
          ' program will automatically infer from file content.'))
@click.option(
    '--filext', '-e', 'input_ext',
    help='Input filename extension following sample ID.')
@click.option(
    '--samples', '-s', type=click.STRING,
    help='List of sample IDs to be included.')
@click.option(
    '--demux/--no-demux', default=None,
    help='Demultiplex alignment by first underscore in query identifier.')
@click.option(
    '--trim-sub', 'trimsub',
    help='Trim subject IDs at the last given delimiter.')
# hierarchies
@click.option(
    '--nodes', 'nodes_fps', type=click.Path(exists=True), multiple=True,
    help='Hierarchies defined by NCBI nodes.dmp or compatible formats.')
@click.option(
    '--newick', 'newick_fps', type=click.Path(exists=True), multiple=True,
    help='Hierarchies defined by a tree in Newick format.')
@click.option(
    '--lineage', 'lineage_fps', type=click.Path(exists=True), multiple=True,
    help='Lineage strings. Can accept Greengenes-style rank prefix.')
@click.option(
    '--columns', 'columns_fps', type=click.Path(exists=True), multiple=True,
    help='Table of classification units per rank (column).')
@click.option(
    '--map', '-m', 'map_fps', type=click.Path(exists=True), multiple=True,
    help='Mapping of lower classification units to higher ones.')
@click.option(
    '--map-as-rank/--map-no-rank', 'map_rank', default=None,
    help='Extract rank name from map filename.')
@click.option(
    '--names', '-n', 'names_fps', type=click.Path(exists=True), multiple=True,
    help=('Names of classification units as defined by NCBI names.dmp or a '
          'simple map.'))
# assignment
@click.option(
    '--rank', '-r', 'ranks', type=click.STRING,
    help=('Classify sequences at this rank. Enter "none" to directly report '
          'subjects; enter "free" for free-rank classification. Can '
          'specify multiple comma-separated ranks.'))
@click.option(
    '--uniq', is_flag=True,
    help=('One sequence can only be assigned to one classification unit, or '
          'remain unassigned if there is ambiguity. Otherwise, all candidate '
          'units are reported and their counts are normalized.'))
@click.option(
    '--major', type=click.IntRange(51, 99),
    help=('In given-rank classification, use majority rule at this percentage '
          'threshold to determine assignment when there are multiple '
          'candidates.'))
@click.option(
    '--above', is_flag=True,
    help=('In given-rank classification, allow assigning a sequence to '
          'a higher rank if it cannot be assigned to the current rank.'))
@click.option(
    '--subok', is_flag=True,
    help=('In free-rank classification, allow assigning a sequence to its '
          'direct subject, if applicable, before going up in hierarchy.'))
# gene matching
@click.option(
    '--coords', '-c', 'coords_fp', type=click.Path(exists=True),
    help='Reference gene coordinates on genomes.')
@click.option(
    '--overlap', type=click.IntRange(1, 100), default=80, show_default=True,
    help='Read/gene overlapping percentage threshold.')
# stratification
@click.option(
    '--stratify', '-t', 'strata_dir',
    type=click.Path(exists=True, file_okay=True, dir_okay=True),
    help='Directory of read-to-feature maps for stratification.')
# normalization
@click.option(
    '--sizes', '-z', type=click.Path(exists=True),
    help=('Divide counts by subject sizes. Can provide a mapping file, or '
          'type "." to calculate from gene coordinates.'))
@click.option(
    '--frac', is_flag=True,
    help='Divide counts by total count of each sample (i.e., fractions).')
@click.option(
    '--scale', type=click.STRING,
    help='Scale counts by this factor. Accepts "k", "M" suffixes.')
@click.option(
    '--digits', type=click.IntRange(0, 10),
    help='Round counts to this number of digits after the decimal point.')
# output files
@click.option(
    '--to-biom/--to-tsv', 'output_fmt', default=None,
    help='Output profile format (BIOM or TSV).')
@click.option(
    '--unassigned', is_flag=True,
    help='Report unassigned sequences.')
@click.option(
    '--name-as-id', is_flag=True,
    help='Replace feature IDs with names.')
@click.option(
    '--add-rank', is_flag=True,
    help='Append feature ranks to table.')
@click.option(
    '--add-lineage', is_flag=True,
    help='Append lineage strings to table.')
@click.option(
    '--outmap', '-u', 'outmap_dir',
    type=click.Path(dir_okay=True),
    help='Write read-to-feature maps to this directory.')
@click.option(
    '--zipmap', 'outmap_zip', default='gz',
    type=click.Choice(['none', 'gz', 'bz2', 'xz'], case_sensitive=False),
    help='Compress read-to-feature maps using this algorithm.')
@click.option(
    '--outcov', 'outcov_dir', type=click.Path(dir_okay=True),
    help='Write subject coverage maps to this directory.')
# performance
@click.option(
    '--chunk', type=click.INT, default=None,
    help='Number of alignment lines to read and parse in each chunk.')
@click.option(
    '--cache', type=click.INT, default=1024,
    help='Number of recent results to cache for faster classification.')
@click.option(
    '--no-exe', is_flag=True,
    help='Disable calling external programs for decompression.')
def classify_cmd(**kwargs):
    """Generate a profile of samples based on a classification system.
    """
    workflow(**kwargs)


# `tools` provides utilities for working with alignments, maps and profiles

@cli.group('tools', cls=NaturalOrderGroup)
def tools():
    """Utilities for working with alignments, maps and profiles.
    """
    pass  # pragma: no cover


@tools.command('normalize')
@click.option(
    '--input', '-i', 'input_fp', required=True,
    type=click.Path(exists=True, dir_okay=False),
    help='Path to input profile.')
@click.option(
    '--output', '-o', 'output_fp', required=True,
    type=click.Path(writable=True, dir_okay=False),
    help='Path to output profile.')
@click.option(
    '--sizes', '-z', 'sizes_fp',
    type=click.Path(exists=True, dir_okay=False),
    help=('Path to mapping of feature sizes, by which values will be divided. '
          'If omitted, will divide values by sum per sample.'))
@click.option(
    '--scale', '-s', type=click.STRING,
    help='Scale values by this factor. Accepts "k", "M" suffixes.')
@click.option(
    '--digits', '-d', type=click.IntRange(0, 10),
    help=('Round values to this number of digits after the decimal point. If '
          'omitted, will keep decimal precision of input profile.'))
@click.pass_context
def normalize_cmd(ctx, **kwargs):
    """Normalize a profile to fractions or by feature sizes.
    """
    normalize_wf(**kwargs)


@tools.command('filter')
@click.option(
    '--input', '-i', 'input_fp', required=True,
    type=click.Path(exists=True, dir_okay=False),
    help='Path to input profile.')
@click.option(
    '--output', '-o', 'output_fp', required=True,
    type=click.Path(writable=True, dir_okay=False),
    help='Path to output profile.')
@click.option(
    '--min-count', '-c', type=click.IntRange(min=1),
    help='Per-sample minimum count threshold.')
@click.option(
    '--min-percent', '-p', type=click.FLOAT,
    help='Per-sample minimum percentage threshold.')
@click.pass_context
def filter_cmd(ctx, **kwargs):
    """Filter a profile by per-sample abundance.
    """
    filter_wf(**kwargs)


@tools.command('merge')
@click.option(
    '--input', '-i', 'input_fps', required=True, multiple=True,
    type=click.Path(exists=True),
    help=('Path to input profiles or directories containing profiles. Can '
          'accept multiple paths.'))
@click.option(
    '--output', '-o', 'output_fp', required=True,
    type=click.Path(writable=True, dir_okay=False),
    help='Path to output profile.')
@click.pass_context
def merge_cmd(ctx, **kwargs):
    """Merge multiple profiles into one profile.
    """
    merge_wf(**kwargs)


@tools.command('collapse')
@click.option(
    '--input', '-i', 'input_fp', required=True,
    type=click.Path(exists=True, dir_okay=False),
    help='Path to input profile.')
@click.option(
    '--map', '-m', 'map_fp', required=True,
    type=click.Path(exists=True, dir_okay=False),
    help=('Mapping of source features to target features. (supports '
          'many-to-many relationships).'))
@click.option(
    '--output', '-o', 'output_fp', required=True,
    type=click.Path(writable=True, dir_okay=False),
    help='Path to output profile.')
@click.option(
    '--divide', '-d', is_flag=True,
    help=('Count each target feature as 1/k (k is the number of targets '
          'mapped to a source). Otherwise, count as one.'))
@click.option(
    '--field', '-f', type=click.INT,
    help='Index of field to be collapsed in a stratified profile.')
@click.option(
    '--names', '-n', 'names_fp', type=click.Path(exists=True),
    help='Names of target features to append to the output profile.')
@click.pass_context
def collapse_cmd(ctx, **kwargs):
    """Collapse a profile based on feature mapping.
    """
    collapse_wf(**kwargs)


@tools.command('coverage')
@click.option(
    '--input', '-i', 'input_fp', required=True,
    type=click.Path(exists=True, dir_okay=False),
    help='Path to input profile.')
@click.option(
    '--map', '-m', 'map_fp', required=True,
    type=click.Path(exists=True, dir_okay=False),
    help='Mapping of feature groups to member features.')
@click.option(
    '--output', '-o', 'output_fp', required=True,
    type=click.Path(writable=True, dir_okay=False),
    help='Path to output coverage table.')
@click.option(
    '--threshold', '-t', type=click.IntRange(1, 100),
    help=('Convert coverage to presence (1) / absence (0) data by this '
          'percentage threshold.'))
@click.option(
    '--count', '-c', is_flag=True,
    help=('Record numbers of covered features instead of percentages '
          '(overrides threshold).'))
@click.option(
    '--names', '-n', 'names_fp', type=click.Path(exists=True),
    help='Names of feature groups to append to the coverage table.')
@click.pass_context
def coverage_cmd(ctx, **kwargs):
    """Calculate per-sample coverage of feature groups.
    """
    coverage_wf(**kwargs)


if __name__ == '__main__':
    cli()  # pragma: no cover



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from unittest import TestCase, main
from os import remove, makedirs
from os.path import join, dirname, realpath
from shutil import rmtree
from tempfile import mkdtemp
import gzip
import bz2

from woltka.file import (
    openzip, readzip, file2stem, path2stem, stem2rank, read_ids,
    id2file_from_dir, id2file_from_map, read_map_uniq, read_map_1st,
    read_map_all, read_map_many, write_readmap)


class FileTests(TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()
        self.datdir = join(dirname(realpath(__file__)), 'data')

    def tearDown(self):
        rmtree(self.tmpdir)

    def test_openzip(self):
        text = 'Hello World!'

        # read regular file
        fp = join(self.tmpdir, 'test.txt')
        with open(fp, 'w') as f:
            f.write(text)
        with openzip(fp) as f:
            obs = f.read()
        self.assertEqual(obs, text)

        # write regular file
        with openzip(fp, 'wt') as f:
            f.write('Here I am!')
        with open(fp, 'r') as f:
            obs = f.read()
        self.assertEqual(obs, 'Here I am!')
        remove(fp)

        # read compressed file
        fpz = join(self.tmpdir, 'test.txt.gz')
        with gzip.open(fpz, 'wb') as f:
            f.write(text.encode())
        with openzip(fpz) as f:
            obs = f.read()
        self.assertEqual(obs, text)
        remove(fpz)

        # write compressed file
        fpb = join(self.tmpdir, 'test.txt.bz2')
        with openzip(fpb, 'at') as f:
            f.write('Here I am!')
        with bz2.open(fpb, 'rt') as f:
            obs = f.read()
        self.assertEqual(obs, 'Here I am!')
        remove(fpb)

    def test_readzip(self):
        text = 'Hello World!'

        # read regular file
        fp = join(self.tmpdir, 'test.txt')
        with open(fp, 'w') as f:
            f.write(text)
        with readzip(fp) as f:
            obs = f.read()
        self.assertEqual(obs, text)

        # read gzip file using builti-in Python module
        fpz = join(self.tmpdir, 'test.txt.gz')
        with gzip.open(fpz, 'wb') as f:
            f.write(text.encode())
        with readzip(fpz) as f:
            obs = f.read()
        self.assertEqual(obs, text)

        # read gzip file using auto-identified gzip program
        zippers = {}
        with readzip(fpz, zippers) as f:
            obs = f.read()
        self.assertEqual(obs, text)
        self.assertDictEqual(zippers, {'gzip': True})

        # read gzip file using already-identified gzip program
        with readzip(fpz, zippers) as f:
            obs = f.read()
        self.assertEqual(obs, text)
        self.assertTrue(zippers['gzip'])

        # disable gzip program
        zippers['gzip'] = False
        with readzip(fpz, zippers) as f:
            obs = f.read()
        self.assertEqual(obs, text)
        self.assertFalse(zippers['gzip'])
        remove(fpz)

    def test_file2stem(self):
        self.assertEqual(file2stem('input.txt'), 'input')
        self.assertEqual(file2stem('input.gz'), 'input')
        self.assertEqual(file2stem('input.txt.gz'), 'input')
        self.assertEqual(file2stem('input.txt.gz', ext='.gz'), 'input.txt')
        with self.assertRaises(ValueError) as ctx:
            file2stem('input.txt', ext='.gz')
        self.assertEqual(str(ctx.exception), (
            'Filepath and filename extension do not match.'))

    def test_path2stem(self):
        self.assertEqual(path2stem('input.txt.bz2'), 'input')
        self.assertEqual(path2stem('/home/input.txt'), 'input')
        self.assertEqual(path2stem('/home/input.fa', ext='a'), 'input.f')
        self.assertEqual(path2stem('/home/.bashrc'), '.bashrc')

    def test_stem2rank(self):
        self.assertEqual(stem2rank('/path/to/db'), 'db')
        self.assertEqual(stem2rank('gene.map.gz'), 'gene')
        self.assertEqual(stem2rank('accession2taxid.txt.bz2'), 'taxid')
        self.assertEqual(stem2rank('compound_2_reaction.txt'), 'reaction')
        self.assertEqual(stem2rank('strain-to-species.map'), 'species')
        self.assertEqual(stem2rank('you_and_i.txt'), 'you_and_i')

    def test_read_ids(self):
        # simple list
        ids = ['a', 'b', 'c']
        obs = read_ids(ids)
        self.assertListEqual(obs, ids)

        # metadata table
        lines = ['#SampleID\tHeight\tWeight',
                 'S01\t120.5\t56.8',
                 'S02\t114.2\t52.5',
                 'S03\t117.8\t55.0',
                 '']
        obs = read_ids(lines)
        self.assertListEqual(obs, ['S01', 'S02', 'S03'])

        # no Id
        with self.assertRaises(ValueError) as ctx:
            read_ids(['#'])
        self.assertEqual(str(ctx.exception), 'No ID is read.')

        # duplicate Ids
        with self.assertRaises(ValueError) as ctx:
            read_ids(['S01', 'S02', 'S01', 'S03'])
        self.assertEqual(str(ctx.exception), 'Duplicate IDs found.')

        # nothing
        self.assertIsNone(read_ids(None))

    def test_id2file_from_dir(self):
        ids = ['a', 'b', 'c']

        # regular Fasta files
        for id_ in ids:
            open(join(self.tmpdir, '{}.faa'.format(id_)), 'a').close()
        obs = id2file_from_dir(self.tmpdir)
        exp = {x: '{}.faa'.format(x) for x in ids}
        self.assertDictEqual(obs, exp)

        # skip subdirectory and only consider files
        sdir = join(self.tmpdir, 'im_dir')
        makedirs(sdir)
        obs = id2file_from_dir(self.tmpdir)
        self.assertDictEqual(obs, exp)
        rmtree(sdir)
        for id_ in ids:
            remove(join(self.tmpdir, '{}.faa'.format(id_)))

        # gzipped Fasta files
        for id_ in ids:
            open(join(self.tmpdir, '{}.faa.gz'.format(id_)), 'a').close()
        obs = id2file_from_dir(self.tmpdir)
        exp = {x: '{}.faa.gz'.format(x) for x in ids}
        self.assertDictEqual(obs, exp)
        for id_ in ids:
            remove(join(self.tmpdir, '{}.faa.gz'.format(id_)))

        # user-defined extension filename
        for id_ in ids:
            open(join(self.tmpdir, '{}.faa'.format(id_)), 'a').close()
        open(join(self.tmpdir, 'readme.txt'), 'a').close()
        open(join(self.tmpdir, 'taxdump'), 'a').close()
        obs = id2file_from_dir(self.tmpdir, ext='.faa')
        exp = {x: '{}.faa'.format(x) for x in ids}
        self.assertDictEqual(obs, exp)
        remove(join(self.tmpdir, 'readme.txt'))
        remove(join(self.tmpdir, 'taxdump'))

        # user-defined Id list
        obs = id2file_from_dir(self.tmpdir, ids=['a', 'b'])
        exp = {x: '{}.faa'.format(x) for x in ['a', 'b']}
        self.assertDictEqual(obs, exp)
        for id_ in ids:
            remove(join(self.tmpdir, '{}.faa'.format(id_)))

        # duplicated Ids
        open(join(self.tmpdir, 'x.faa'), 'a').close()
        open(join(self.tmpdir, 'x.tsv'), 'a').close()
        with self.assertRaises(ValueError) as ctx:
            id2file_from_dir(self.tmpdir)
        msg = 'Ambiguous files for ID: "x".'
        self.assertEqual(str(ctx.exception), msg)
        remove(join(self.tmpdir, 'x.faa'))
        remove(join(self.tmpdir, 'x.tsv'))

        # real files
        dir_ = join(self.datdir, 'align', 'bowtie2')

        obs = id2file_from_dir(dir_)
        exp = {f'S0{i}': f'S0{i}.sam.xz' for i in range(1, 6)}
        self.assertDictEqual(obs, exp)

        obs = id2file_from_dir(dir_, ext='.xz')
        exp = {f'S0{i}.sam': f'S0{i}.sam.xz' for i in range(1, 6)}
        self.assertDictEqual(obs, exp)

        obs = id2file_from_dir(dir_, ids=['S01', 'S02', 'S03'])
        exp = {f'S0{i}': f'S0{i}.sam.xz' for i in range(1, 4)}
        self.assertDictEqual(obs, exp)

    def test_id2file_from_map(self):
        # mock alignment files
        ids = ['a', 'b', 'c']
        fps = [join(self.tmpdir, '{}.fq'.format(x)) for x in ids]
        for fp in fps:
            open(fp, 'a').close()

        # mapping file with absolute paths
        mapfile = join(self.tmpdir, 'map.tmp')
        with open(mapfile, 'w') as f:
            for id_, fp in zip(ids, fps):
                f.write('{}\t{}\n'.format(id_, fp))
        obs = id2file_from_map(mapfile)
        exp = [pair for pair in zip(ids, fps)]
        self.assertListEqual(obs, exp)

        # relative paths to same directory
        with open(mapfile, 'w') as f:
            f.write('# This is a mapping file!\n\n')
            for id_ in ids:
                f.write('{}\t{}.fq\n'.format(id_, id_))
        obs = id2file_from_map(mapfile)
        self.assertListEqual(obs, exp)

        # not a mapping file
        with open(mapfile, 'w') as f:
            f.write('This is NOT a mapping file!\n\n')
        obs = id2file_from_map(mapfile)
        self.assertIsNone(obs)

        # incorrect filepaths starting from 1st line
        with open(mapfile, 'w') as f:
            for id_ in ids:
                f.write('{}\t{}.pdf\n'.format(id_, id_))
        obs = id2file_from_map(mapfile)
        self.assertIsNone(obs)

        # incorrect filepath(s) in middle of file
        with open(mapfile, 'w') as f:
            f.write('a\ta.fq\n')
            f.write('b\tb.fq\n')
            f.write('d\td.fq\n')
        with self.assertRaises(ValueError) as ctx:
            id2file_from_map(mapfile)
        self.assertEqual(str(ctx.exception), (
            'Alignment file "d.fq" does not exist.'))

        # empty mapping file
        open(mapfile, 'w').close()
        obs = id2file_from_map(mapfile)
        self.assertIsNone(obs)
        for fp in fps:
            remove(fp)

        # real files
        dir_ = join(self.datdir, 'align', 'bowtie2')
        exp = []
        with open(mapfile, 'w') as f:
            for i in range(1, 6):
                id_ = f'S0{i}'
                fp = join(dir_, f'{id_}.sam.xz')
                f.write(f'{id_}\t{fp}\n')
                exp.append((id_, fp))
        obs = id2file_from_map(mapfile)
        self.assertListEqual(obs, exp)
        remove(mapfile)

    def test_read_map_uniq(self):
        # simplest map (two columns)
        tsv = ('R1	A',
               'R2	B',
               'R3	C')
        exp = (('R1', 'A'), ('R2', 'B'), ('R3', 'C'))
        for obs_, exp_ in zip(read_map_uniq(tsv), exp):
            self.assertTupleEqual(obs_, exp_)

        # specify separator
        obs = read_map_uniq(('R1,A',), sep=',')
        self.assertTupleEqual(next(obs), ('R1', 'A'))

        # single column
        obs = tuple(read_map_uniq(('Hello.',)))
        self.assertTupleEqual(obs, ())

        # more than two columns
        obs = tuple(read_map_uniq(('Here\tit\tis.',)))
        self.assertTupleEqual(obs, ())

    def test_read_map_1st(self):
        # simplest map (two columns)
        tsv = ('R1	A',
               'R2	B',
               'R3	C')
        exp = (('R1', 'A'), ('R2', 'B'), ('R3', 'C'))
        for obs_, exp_ in zip(read_map_1st(tsv), exp):
            self.assertTupleEqual(obs_, exp_)

        # specify separator
        obs = read_map_1st(('R1,A',), sep=',')
        self.assertTupleEqual(next(obs), ('R1', 'A'))

        # single column
        obs = tuple(read_map_1st(('Hello.',)))
        self.assertTupleEqual(obs, ())

        # more than two columns
        obs = read_map_1st(('Here\tit\tis.',))
        self.assertTupleEqual(next(obs), ('Here', 'it'))

    def test_read_map_all(self):
        # simple map (two or more columns)
        tsv = ('R1	A',
               'R2	A	B',
               'R3	C	A	B')
        exp = (('R1', ['A']), ('R2', ['A', 'B']), ('R3', ['C', 'A', 'B']))
        for obs_, exp_ in zip(read_map_all(tsv), exp):
            self.assertTupleEqual(obs_, exp_)

        # specify separator
        obs = read_map_all(('R1,A,B,C',), sep=',')
        self.assertTupleEqual(next(obs), ('R1', ['A', 'B', 'C']))

        # single column
        obs = tuple(read_map_all(('Hello.',)))
        self.assertTupleEqual(obs, ())

    def test_read_map_many(self):
        # one-to-one
        tsv = ('R1	A',
               'R2	B',
               'R3	C')
        obs = read_map_many(tsv)
        exp = {'R1': ['A'], 'R2': ['B'], 'R3': ['C']}
        self.assertDictEqual(obs, exp)

        # many-to-one, with separator
        tsv = ('R1,A',
               'R2,B',
               'R3,B',
               'R4,C')
        obs = read_map_many(tsv, sep=',')
        exp = {'R1': ['A'], 'R2': ['B'], 'R3': ['B'], 'R4': ['C']}
        self.assertDictEqual(obs, exp)

        # one-to-many
        tsv = ('R1	A',
               'R2	A	B',
               'R3	B	C	D',
               'R4	A	D')
        obs = read_map_many(tsv)
        exp = {'R1': ['A'], 'R2': ['A', 'B'], 'R3': ['B', 'C', 'D'],
               'R4': ['A', 'D']}
        self.assertDictEqual(obs, exp)

        # many-to-many
        tsv = ('R1	A',
               'R1	B',
               'R2	B',
               'R3	A',
               'R3	C',
               'R4	C')
        obs = read_map_many(tsv)
        exp = {'R1': ['A', 'B'], 'R2': ['B'], 'R3': ['A', 'C'], 'R4': ['C']}
        self.assertDictEqual(obs, exp)

        # no relationship
        tsv = ('Hello.')
        obs = read_map_many(tsv)
        self.assertDictEqual(obs, {})

    def test_write_readmap(self):
        # typical read map
        rmap = {'R1': 'G1',
                'R2': 'G2',
                'R3': ['G2', 'G1', 'G2'],
                'R4': ['G3', 'G3', 'G3'],
                'R5': None}
        fp = join(self.tmpdir, 'readmap.tsv')
        with open(fp, 'w') as f:
            write_readmap(f, rmap.keys(), rmap.values())
        with open(fp, 'r') as f:
            obs = sorted(f.read().splitlines())
        exp = ['R1\tG1', 'R2\tG2', 'R3\tG2:2\tG1:1', 'R4\tG3:3']
        self.assertListEqual(obs, exp)

        # with name dict
        namedic = {'G1': 'Ecoli', 'G2': 'Strep', 'G3': 'Kleb'}
        with open(fp, 'w') as f:
            write_readmap(f, rmap.keys(), rmap.values(), namedic)
        with open(fp, 'r') as f:
            obs = sorted(f.read().splitlines())
        exp = ['R1\tEcoli', 'R2\tStrep', 'R3\tStrep:2\tEcoli:1', 'R4\tKleb:3']
        self.assertListEqual(obs, exp)
        remove(fp)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from qiime2.plugin import SemanticType
from q2_types.feature_data import FeatureData


SeqAlnMap = SemanticType('SeqAlnMap', variant_of=FeatureData.field['type'])
BLAST6Out = SemanticType('BLAST6Out', variant_of=FeatureData.field['type'])
SimpleMap = SemanticType('SimpleMap', variant_of=FeatureData.field['type'])
NCBINodes = SemanticType('NCBINodes', variant_of=FeatureData.field['type'])
GeneCoordinates = SemanticType('GeneCoordinates')



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from qiime2.plugin import model


class SeqAlnMapFormat(model.TextFileFormat):
    def _validate_(self, level):
        pass


SeqAlnMapDirFmt = model.SingleFileDirectoryFormat(
    'SeqAlnMapDirFmt', 'alignment.sam', SeqAlnMapFormat)


class BLAST6OutFormat(model.TextFileFormat):
    def _validate_(self, level):
        pass


BLAST6OutDirFmt = model.SingleFileDirectoryFormat(
    'BLAST6OutDirFmt', 'alignment.b6o', BLAST6OutFormat)


class SimpleMapFormat(model.TextFileFormat):
    def _validate_(self, level):
        pass


SimpleMapDirFmt = model.SingleFileDirectoryFormat(
    'SimpleMapDirFmt', 'mapping.txt', SimpleMapFormat)


class NCBINodesFormat(model.TextFileFormat):
    def _validate_(self, level):
        pass


NCBINodesDirFmt = model.SingleFileDirectoryFormat(
    'NCBINodesDirFmt', 'nodes.dmp', NCBINodesFormat)


class GeneCoordFormat(model.TextFileFormat):
    def _validate_(self, level):
        pass


GeneCoordDirFmt = model.SingleFileDirectoryFormat(
    'GeneCoordDirFmt', 'coordinates.txt', GeneCoordFormat)


# class NCBINamesFormat(model.TextFileFormat):
#     def _validate_(self, level):
#         pass


# class NCBITaxdumpDirFmt(model.DirectoryFormat):
#     nodes = model.File('nodes.dmp', format=NCBINodesFormat)
#     names = model.File('names.dmp', format=NCBINamesFormat)



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

"""Unit tests for the ordinal system, as well as a demonstration of how the
system works. See docstrings of `test_match_read_gene_dummy`.
"""

from unittest import TestCase, main
from os.path import join, dirname, realpath
from shutil import rmtree
from tempfile import mkdtemp
from io import StringIO
from functools import partial

from woltka.file import openzip
from woltka.align import parse_b6o_file_ext, parse_sam_file_ext
from woltka.ordinal import (
    match_read_gene_dummy, match_read_gene, match_read_gene_naive,
    match_read_gene_quart, ordinal_parser_dummy, ordinal_mapper,
    load_gene_coords, calc_gene_lens)


class OrdinalTests(TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()
        self.datdir = join(dirname(realpath(__file__)), 'data')

    def tearDown(self):
        rmtree(self.tmpdir)

    def test_match_read_gene_dummy(self):
        """This test demonstrates how genes and reads are matched in an
        ordinal system. It is for demonstration purpose.

        Illustration of gene (>) and read (-) locations on a genome (=):

           reads:             ---------r1---------
                                    ---------r2---------
                                        ---------r3---------
                                          ---------r4---------
                                                  ---------r5---------
                                                  ---------r6---------
          genome:  1 ====>>>>>>>>>>>g1>>>>>>>>>>>>===>>>>>>>>>>>>>>g2>> 50

           reads:             ---------r7---------
                                   ---------r8---------
                                                    ------r9------
          genome: 51 >>>>>>>>>>>===>>>>>>>>>>>>>>g3>>>>>>>>>>>>>======= 100
          --------------------------------------------------

        This function finds matching read/gene pairs given their overlap
        is no less than a fraction of alignment length (default: 80%). In this
        example, the matching pairs are:

          r1 - g1  (whole read within gene)
          r5 - g2  (17 / 20 nt within gene, exceeding threshold)
          r6 - g2  (same as above)
          r8 - g3  (whole read within gene)

        """
        # gene table
        genes = [('g1',  5, 29),
                 ('g2', 33, 61),
                 ('g3', 65, 94)]
        # read map
        reads = [('r1', 10, 29),
                 ('r2', 16, 35),
                 ('r3', 20, 39),
                 ('r4', 22, 41),
                 ('r5', 30, 49),
                 ('r6', 30, 49),  # identical
                 ('r7', 60, 79),
                 ('r8', 65, 84),
                 ('r9', 82, 95)]  # shorter

        # alignment length map
        lens = {'r{}'.format(i): 20 for i in range(1, 10)}

        # flatten lists
        genes = [x for id_, beg, end in genes for x in (
            (beg,  True,  True, id_),
            (end, False,  True, id_))]
        reads = [x for id_, beg, end in reads for x in (
            (beg,  True, False, id_),
            (end, False, False, id_))]

        # merge genes and reads and sort
        queue = sorted(genes + reads)

        # default (threshold = 80%)
        obs = list(match_read_gene_dummy(queue, lens, th=0.8))
        exp = [('r1', 'g1'),
               ('r5', 'g2'),
               ('r6', 'g2'),
               ('r8', 'g3')]
        self.assertListEqual(obs, exp)

        # threashold = 50%
        obs = list(match_read_gene_dummy(queue, lens, th=0.5))
        exp = [('r1', 'g1'),
               ('r2', 'g1'),
               ('r3', 'g1'),
               ('r5', 'g2'),
               ('r6', 'g2'),
               ('r7', 'g3'),
               ('r8', 'g3'),
               ('r9', 'g3')]
        self.assertListEqual(obs, exp)

    def test_match_read_gene(self):
        """This test is for the real function.
        """
        # gene table
        genes = [(1,  5, 29),
                 (2, 33, 61),
                 (3, 65, 94)]
        # read map
        reads = [(1, 10, 29),
                 (2, 16, 35),
                 (3, 20, 39),
                 (4, 22, 41),
                 (5, 30, 49),
                 (6, 30, 49),  # identical
                 (7, 60, 79),
                 (8, 65, 84),
                 (9, 82, 95)]  # shorter

        # read length is uniformly 20, threshold is 80%, so effective
        # alignment length is 20 * 0.8 = 16
        genes = [x for idx, beg, end in genes for x in (
            (beg << 48) + (1 << 31) + (1 << 30) + idx,
            (end << 48) + (0 << 31) + (1 << 30) + idx)]
        reads = [x for idx, beg, end in reads for x in (
            (beg << 48) + (16 << 31) + (0 << 30) + idx,
            (end << 48) + (0 << 31) + (0 << 30) + idx)]
        queue = sorted(genes + reads)

        # default
        obs = list(match_read_gene(queue))
        exp = [(1, 1),
               (5, 2),
               (6, 2),
               (8, 3)]
        self.assertListEqual(obs, exp)

    def test_match_read_gene_naive(self):
        """The naive solution should produce identical result compared to
        the default (ordinal) solution.
        """
        genes = [(1,  5, 29),
                 (2, 33, 61),
                 (3, 65, 94)]
        reads = [(1, 10, 29),
                 (2, 16, 35),
                 (3, 20, 39),
                 (4, 22, 41),
                 (5, 30, 49),
                 (6, 30, 49),
                 (7, 60, 79),
                 (8, 65, 84),
                 (9, 82, 95)]
        genes = [x for idx, beg, end in genes for x in (
            (beg << 48) + (1 << 31) + (1 << 30) + idx,
            (end << 48) + (0 << 31) + (1 << 30) + idx)]
        reads = [x for idx, beg, end in reads for x in (
            (beg << 48) + (14 << 31) + (0 << 30) + idx,
            (end << 48) + (0 << 31) + (0 << 30) + idx)]

        # don't sort, but directly feed both queues
        obs = list(match_read_gene_naive(genes, reads))
        exp = [(1, 1),
               (2, 1),
               (5, 2),
               (6, 2),
               (7, 3),
               (8, 3)]
        self.assertListEqual(obs, exp)

    def test_match_read_gene_quart(self):
        """The naive solution should produce identical result compared to
        the default (ordinal) solution.
        """
        genes = [(1,  5, 29),
                 (2, 33, 61),
                 (3, 65, 94),
                 (4, 61, 76),  # added a small gene within a read
                 (5, 68, 72)]  # added a tiny gene
        reads = [(1, 10, 29),
                 (2, 16, 35),
                 (3, 20, 39),
                 (4, 22, 41),
                 (5, 30, 49),
                 (6, 30, 49),
                 (7, 60, 79),
                 (8, 65, 84),
                 (9, 82, 95)]
        genes = [x for idx, beg, end in genes for x in (
            (beg << 48) + (1 << 31) + (1 << 30) + idx,
            (end << 48) + (0 << 31) + (1 << 30) + idx)]
        genes.sort()
        reads = [x for idx, beg, end in reads for x in (
            (beg << 48) + (14 << 31) + (0 << 30) + idx,
            (end << 48) + (0 << 31) + (0 << 30) + idx)]

        obs = list(match_read_gene_quart(genes, reads))
        exp = [(1, 1),
               (2, 1),
               (5, 2),
               (6, 2),
               (7, 4),  # match to added small gene
               (7, 3),
               (8, 3)]
        self.assertListEqual(obs, exp)

        # a special case with a giant read
        genes = [(1, 1, 5),
                 (2, 6, 7),
                 (3, 7, 8)]
        reads = [(1, 4, 9)]
        genes = [x for idx, beg, end in genes for x in (
            (beg << 48) + (1 << 31) + (1 << 30) + idx,
            (end << 48) + (0 << 31) + (1 << 30) + idx)]
        genes.sort()
        reads = [x for idx, beg, end in reads for x in (
            (beg << 48) + (5 << 31) + (0 << 30) + idx,
            (end << 48) + (0 << 31) + (0 << 30) + idx)]
        obs = list(match_read_gene_quart(genes, reads))
        exp = []
        self.assertListEqual(obs, exp)

    def test_ordinal_parser_dummy(self):
        # b6o (BLAST, DIAMOND, BURST, etc.)
        b6o = (
            'S1/1	NC_123456	100	100	0	0	1	100	225	324	1.2e-30	345',
            'S1/2	NC_123456	95	98	2	1	2	99	708	608	3.4e-20	270')
        obs = ordinal_parser_dummy(b6o, parse_b6o_file_ext)
        self.assertListEqual(obs[0], ['S1/1', 'S1/2'])
        self.assertDictEqual(obs[1], {'NC_123456': {0: 100, 1: 98}})
        self.assertDictEqual(obs[2], {'NC_123456': [
            (225, True, False, 0), (324, False, False, 0),
            (608, True, False, 1), (708, False, False, 1)]})

        # sam (BWA, Bowtie2, Minimap2 etc.)
        sam = iter((
            # SAM header to be ignored
            '@HD	VN:1.0	SO:unsorted',
            # normal, fully-aligned, forward strand
            'S1	77	NC_123456	26	0	100M	*	0	0	*	*',
            # shortened, reverse strand
            'S1	141	NC_123456	151	0	80M	*	0	0	*	*',
            # not perfectly aligned, unpaired
            'S2	0	NC_789012	186	0	50M5I20M5D20M	*	0	0	*	*',
            # unaligned
            'S2	16	*	0	0	*	*	0	0	*	*'))
        obs = ordinal_parser_dummy(sam, parse_sam_file_ext)
        self.assertListEqual(obs[0], ['S1/1', 'S1/2', 'S2'])
        self.assertDictEqual(obs[1], {
            'NC_123456': {0: 100, 1: 80},
            'NC_789012': {2: 90}})
        self.assertDictEqual(obs[2], {
            'NC_123456': [(26,  True, False, 0), (125, False, False, 0),
                          (151, True, False, 1), (230, False, False, 1)],
            'NC_789012': [(186, True, False, 2), (280, False, False, 2)]})

    def test_ordinal_mapper(self):
        # uses the same example as above, with some noises
        coords, idmap, _ = load_gene_coords((
            '>n1',
            'g1	5	29',
            'g2	33	61',
            'g3	65	94',
            'gx	108	135'))
        aln = StringIO('\n'.join((
            'r1	n1	95	20	0	0	1	20	10	29	1	1',
            'r2	n1	95	20	0	0	1	20	16	35	1	1',
            'r3	n1	95	20	0	0	1	20	20	39	1	1',
            'r4	n1	95	20	0	0	20	1	22	41	1	1',
            'r5	n1	95	20	0	0	20	1	30	49	1	1',
            'rx	nx	95	20	0	0	1	20	1	20	1	1',
            'r6	n1	95	20	0	0	1	20	49	30	1	1',
            'r7	n1	95	20	0	0	25	6	79	60	1	1',
            'r8	n1	95	20	0	0	1	20	84	65	1	1',
            'r9	n1	95	20	0	0	1	20	95	82	1	1',
            'rx	nx	95	0	0	0	0	0	0	0	1	1',
            '# end of file')))
        obs = list(ordinal_mapper(aln, coords, idmap))[0]
        exp = [('r1', 'g1'),
               ('r5', 'g2'),
               ('r6', 'g2'),
               ('r8', 'g3')]
        self.assertListEqual(list(obs[0]), [x[0] for x in exp])
        self.assertListEqual(list(obs[1]), [{x[1]} for x in exp])

        # specify format
        aln.seek(0)
        obs = list(ordinal_mapper(aln, coords, idmap, fmt='b6o'))[0]
        self.assertListEqual(list(obs[0]), [x[0] for x in exp])
        self.assertListEqual(list(obs[1]), [{x[1]} for x in exp])

        # specify chunk size
        aln.seek(0)
        obs = list(ordinal_mapper(aln, coords, idmap, n=5))
        self.assertListEqual(list(obs[0][0]), [x[0] for x in exp[:2]])
        self.assertListEqual(list(obs[0][1]), [{x[1]} for x in exp[:2]])
        self.assertListEqual(list(obs[1][0]), [x[0] for x in exp[2:]])
        self.assertListEqual(list(obs[1][1]), [{x[1]} for x in exp[2:]])

        # add prefix
        aln.seek(0)
        obs = list(ordinal_mapper(aln, coords, idmap, prefix=True))[0]
        self.assertListEqual(list(obs[0]), [x[0] for x in exp])
        self.assertListEqual(list(obs[1]), [{f'n1_{x[1]}'} for x in exp])

        # specify threshold
        aln.seek(0)
        obs = list(ordinal_mapper(aln, coords, idmap, th=0.5))[0]
        exp = [('r1', 'g1'),
               ('r2', 'g1'),
               ('r3', 'g1'),
               ('r5', 'g2'),
               ('r6', 'g2'),
               ('r7', 'g3'),
               ('r8', 'g3'),
               ('r9', 'g3')]
        self.assertListEqual(list(obs[0]), [x[0] for x in exp])
        self.assertListEqual(list(obs[1]), [{x[1]} for x in exp])

    def test_load_gene_coords(self):
        # simple case
        tbl = ('>n1',
               'g1	5	29',
               'g2	33	61',
               'g3	65	94')
        obs, idmap, isdup = load_gene_coords(tbl)
        self.assertFalse(isdup)
        self.assertDictEqual(idmap, {'n1': ['g1', 'g2', 'g3']})
        exp = {'n1': [
            (5 << 48) + (3 << 30) + 0,
            (29 << 48) + (1 << 30) + 0,
            (33 << 48) + (3 << 30) + 1,
            (61 << 48) + (1 << 30) + 1,
            (65 << 48) + (3 << 30) + 2,
            (94 << 48) + (1 << 30) + 2]}
        self.assertDictEqual(obs, exp)

        # NCBI accession
        tbl = ('## GCF_000123456\n',
               '# NC_123456\n',
               '1	5	384\n',
               '2	410	933\n',
               '# NC_789012\n',
               '1	912	638\n',
               '2	529	75\n')
        obs, idmap, isdup = load_gene_coords(tbl, sort=True)
        self.assertTrue(isdup)
        self.assertDictEqual(idmap, {
            'NC_123456': ['1', '2'], 'NC_789012': ['1', '2']})
        exp = {'NC_123456': [
            (5 << 48) + (3 << 30) + 0,
            (384 << 48) + (1 << 30) + 0,
            (410 << 48) + (3 << 30) + 1,
            (933 << 48) + (1 << 30) + 1],
               'NC_789012': [
            (75 << 48) + (3 << 30) + 1,
            (529 << 48) + (1 << 30) + 1,
            (638 << 48) + (3 << 30) + 0,
            (912 << 48) + (1 << 30) + 0]}
        self.assertDictEqual(obs, exp)

        # don't sort
        obs = load_gene_coords(tbl, sort=False)[0]['NC_789012']
        exp = [(638 << 48) + (3 << 30) + 0, (912 << 48) + (1 << 30) + 0,
               (75 << 48) + (3 << 30) + 1,  (529 << 48) + (1 << 30) + 1]
        self.assertListEqual(obs, exp)

        # incorrect formats
        # only one column
        msg = 'Cannot extract coordinates from line:'
        with self.assertRaises(ValueError) as ctx:
            load_gene_coords(('hello',))
        self.assertEqual(str(ctx.exception), f'{msg} "hello".')
        # only two columns
        with self.assertRaises(ValueError) as ctx:
            load_gene_coords(('hello\t100',))
        self.assertEqual(str(ctx.exception), f'{msg} "hello\t100".')
        # three columns but 3rd is string
        with self.assertRaises(ValueError) as ctx:
            load_gene_coords(('hello\t100\tthere',))
        self.assertEqual(str(ctx.exception), f'{msg} "hello\t100\tthere".')

        # real coords file
        fp = join(self.datdir, 'function', 'coords.txt.xz')
        with openzip(fp) as f:
            obs, idmap, isdup = load_gene_coords(f, sort=True)
        self.assertTrue(isdup)
        self.assertEqual(len(idmap), 107)
        self.assertEqual(len(obs), 107)
        obs_ = obs['G000006745']
        self.assertEqual(len(obs_), 7188)
        self.assertEqual(obs_[0], (372 << 48) + (3 << 30) + 0)
        self.assertEqual(obs_[1], (806 << 48) + (1 << 30) + 0)
        self.assertEqual(obs_[2], (816 << 48) + (3 << 30) + 1)
        self.assertEqual(obs_[3], (2177 << 48) + (1 << 30) + 1)

    def test_calc_gene_lens(self):
        coords = {'NC_123456': [(5 << 48) + (3 << 30) + 0,
                                (384 << 48) + (1 << 30) + 0,
                                (410 << 48) + (3 << 30) + 1,
                                (933 << 48) + (1 << 30) + 1],
                  'NC_789012': [(75 << 48) + (3 << 30) + 1,
                                (529 << 48) + (1 << 30) + 1,
                                (638 << 48) + (3 << 30) + 0,
                                (912 << 48) + (1 << 30) + 0]}
        idmap = {'NC_123456': ['1', '2'],
                 'NC_789012': ['1', '2']}
        mapper = partial(ordinal_mapper, coords=coords, idmap=idmap,
                         prefix=True)
        obs = calc_gene_lens(mapper)
        exp = {'NC_123456_1': 380,
               'NC_123456_2': 524,
               'NC_789012_2': 455,
               'NC_789012_1': 275}
        self.assertDictEqual(obs, exp)

        idmap = {'NC_123456': ['NP_135792.1', 'NP_246801.2'],
                 'NC_789012': ['NP_258147.1', 'NP_369258.2']}
        mapper = partial(ordinal_mapper, coords=coords, idmap=idmap,
                         prefix=False)
        obs = calc_gene_lens(mapper)
        exp = {'NP_135792.1': 380,
               'NP_246801.2': 524,
               'NP_369258.2': 455,
               'NP_258147.1': 275}
        self.assertDictEqual(obs, exp)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from unittest import TestCase, main
from os import remove, makedirs
from os.path import join, dirname, realpath
from shutil import rmtree, copy
from tempfile import mkdtemp

from woltka.tools import (
    normalize_wf, filter_wf, merge_wf, collapse_wf, coverage_wf)


class ToolsTests(TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()
        self.datdir = join(dirname(realpath(__file__)), 'data')

    def tearDown(self):
        rmtree(self.tmpdir)

    def test_normalize_wf(self):
        input_fp = join(self.datdir, 'output', 'bowtie2.ogu.tsv')
        output_fp = join(self.tmpdir, 'tmp.tsv')

        # relative abundance
        normalize_wf(input_fp, output_fp, digits=3)
        with open(output_fp, 'r') as f:
            obs = f.read().splitlines()
        self.assertIn('G000007325\t0.0\t0.013\t0.0\t0.0\t0.113', obs)
        self.assertIn('G000215745\t0.0\t0.0\t0.0\t0.944\t0.0', obs)

        # by genome size (cpm)
        sizes_fp = join(self.datdir, 'taxonomy', 'length.map')
        normalize_wf(input_fp, output_fp, sizes_fp, scale='1M')
        with open(output_fp, 'r') as f:
            obs = f.read().splitlines()
        self.assertIn('G000007325\t0\t12\t0\t0\t104', obs)
        self.assertIn('G000215745\t0\t0\t0\t358\t0', obs)

        # bad scale param
        with self.assertRaises(SystemExit) as ctx:
            normalize_wf(input_fp, output_fp, scale='Hi!')
        errmsg = '"Hi!" is not a valid scale factor.'
        self.assertEqual(str(ctx.exception), errmsg)

        # missing sizes
        sizes_fp = join(self.datdir, 'function', 'coords.txt.xz')
        with self.assertRaises(SystemExit) as ctx:
            normalize_wf(input_fp, output_fp, sizes_fp)
        errmsg = 'One or more features are not found in the size map.'
        self.assertEqual(str(ctx.exception), errmsg)
        remove(output_fp)

    def test_filter_wf(self):
        input_fp = join(self.datdir, 'output', 'blastn.species.tsv')
        output_fp = join(self.tmpdir, 'tmp.tsv')

        # filter by count
        filter_wf(input_fp, output_fp, min_count=5)
        with open(output_fp, 'r') as f:
            n = len(f.read().splitlines())
        self.assertEqual(n, 44)

        # filter by percentage
        filter_wf(input_fp, output_fp, min_percent=1)
        with open(output_fp, 'r') as f:
            n = len(f.read().splitlines())
        self.assertEqual(n, 31)
        remove(output_fp)

        # no threshold specified
        with self.assertRaises(SystemExit) as ctx:
            filter_wf(input_fp, output_fp)
        errmsg = ('Please specify either minimum count or minimum percentage '
                  'threshold.')
        self.assertEqual(str(ctx.exception), errmsg)

        # zero is not a valid threshold
        with self.assertRaises(SystemExit) as ctx:
            filter_wf(input_fp, output_fp, min_count=0, min_percent=0)
        self.assertEqual(str(ctx.exception), errmsg)

        # both thresholds specified
        with self.assertRaises(SystemExit) as ctx:
            filter_wf(input_fp, output_fp, min_count=10, min_percent=10)
        errmsg = ('Only one of minimum count or minimum percentage thresholds '
                  'can be specified.')
        self.assertEqual(str(ctx.exception), errmsg)

        # percentage threshold too large
        with self.assertRaises(SystemExit) as ctx:
            filter_wf(input_fp, output_fp, min_percent=120)
        errmsg = 'Minimum percentage threshold must be below 100.'
        self.assertEqual(str(ctx.exception), errmsg)

    def test_merge_wf(self):
        # files
        input_fp1 = join(self.datdir, 'output', 'burst.process.tsv')
        input_fp2 = join(self.datdir, 'output', 'split.process.tsv')
        output_fp = join(self.tmpdir, 'tmp.tsv')
        merge_wf([input_fp1, input_fp2], output_fp)
        with open(output_fp, 'r') as f:
            n = len(f.read().splitlines())
        self.assertEqual(n, 205)
        remove(output_fp)

        # directory
        input_fp = join(self.tmpdir, 'indir')
        makedirs(input_fp)

        # only one profile
        copy(input_fp1, join(input_fp, 'test1.tsv'))
        with self.assertRaises(SystemExit) as ctx:
            merge_wf([input_fp], output_fp)
        errmsg = 'Please provide two or more profiles.'
        self.assertEqual(str(ctx.exception), errmsg)

        # two profiles
        copy(input_fp2, join(input_fp, 'test2.tsv'))
        merge_wf([input_fp], output_fp)
        with open(output_fp, 'r') as f:
            n = len(f.read().splitlines())
        self.assertEqual(n, 205)
        remove(output_fp)

        # invalid profile
        copy(join(self.datdir, 'tree.nwk'), join(input_fp, 'test3.tsv'))
        with self.assertRaises(SystemExit) as ctx:
            merge_wf([input_fp], output_fp)
        errmsg = 'Cannot parse test3.tsv as a profile.'
        self.assertEqual(str(ctx.exception), errmsg)

        # keep original decimal precision
        input_fp1 = join(self.tmpdir, '1st.tsv')
        with open(input_fp1, 'w') as f:
            f.write('#FeatureID\tS01\nG01\t1.1')
        input_fp2 = join(self.tmpdir, '2nd.tsv')
        with open(input_fp2, 'w') as f:
            f.write('#FeatureID\tS01\nG01\t2.2')
        output_fp = join(self.tmpdir, 'tmp.tsv')
        merge_wf([input_fp1, input_fp2], output_fp)
        with open(output_fp, 'r') as f:
            obs = f.read().splitlines()
        exp = ['#FeatureID\tS01', 'G01\t3.3']
        self.assertListEqual(obs, exp)
        remove(output_fp)
        remove(input_fp1)
        remove(input_fp2)

    def test_collapse_wf(self):
        input_fp = join(self.datdir, 'output', 'truth.gene.tsv')
        map_fp = join(self.datdir, 'function', 'nucl', 'uniref.map.xz')
        output_fp = join(self.tmpdir, 'tmp.tsv')

        # one-to-one
        collapse_wf(input_fp, map_fp, output_fp)
        with open(output_fp, 'r') as f:
            obs = f.read().splitlines()
        self.assertEqual(len(obs), 2489)
        for line in obs:
            if line.startswith('A0A069B1G8'):
                self.assertEqual(line[11:], '1\t0\t0\t0\t0')
            if line.startswith('P0AGG9\t'):
                self.assertEqual(line[7:], '0\t0\t0\t2\t0')

        # one-to-many
        input_fp = join(self.datdir, 'output', 'truth.uniref.tsv')
        map_fp = join(self.datdir, 'function', 'go', 'goslim.tsv.xz')
        names_fp = join(self.datdir, 'function', 'go', 'name.txt.xz')
        collapse_wf(input_fp, map_fp, output_fp, names_fp=names_fp)
        with open(output_fp, 'r') as f:
            obs = f.read().splitlines()
        self.assertEqual(len(obs), 84)
        for line in obs:
            if line.startswith('GO:0005737'):
                self.assertEqual(line[11:], '59\t55\t62\t22\t47\tcytoplasm')
            if line.startswith('GO:0004518'):
                self.assertEqual(line[11:], '3\t7\t9\t1\t4\tnuclease activity')
        remove(output_fp)

        # wrong mapping file
        map_fp = join(self.datdir, 'tree.nwk')
        with self.assertRaises(SystemExit) as ctx:
            collapse_wf(input_fp, map_fp, output_fp, divide=True)
        errmsg = 'No source-target relationship is found in tree.nwk.'
        self.assertEqual(str(ctx.exception), errmsg)

        # stratified profile
        input_fp = join(self.datdir, 'output', 'burst.genus.process.tsv')
        map_fp = join(self.datdir, 'function', 'go', 'go2slim.map.xz')
        collapse_wf(input_fp, map_fp, output_fp, field=2)
        with open(output_fp, 'r') as f:
            obs = f.read().splitlines()
        self.assertEqual(len(obs), 279)
        for line in obs:
            if line.startswith('Klebsiella|GO:0008150'):
                self.assertEqual(line[22:], '2\t47\t0\t87\t0')
            if line.startswith('Streptococcus|GO:0008150'):
                self.assertEqual(line[25:], '0\t2\t9\t3\t0')
        remove(output_fp)

    def test_coverage_wf(self):
        input_fp = join(self.datdir, 'output', 'truth.metacyc.tsv')
        map_fp = join(self.datdir, 'function', 'metacyc', 'pathway_mbrs.txt')
        output_fp = join(self.tmpdir, 'tmp.tsv')

        # default behavior
        coverage_wf(input_fp, map_fp, output_fp)
        with open(output_fp, 'r') as f:
            obs = f.read().splitlines()
        self.assertEqual(len(obs), 316)
        for line in obs:
            if line.startswith('ARABCAT-PWY\t'):
                self.assertListEqual(line.split('\t')[1:], [
                    '20.0', '80.0', '100.0', '80.0', '100.0'])

        # parameters
        names_fp = join(self.datdir, 'function', 'metacyc', 'pathway_name.txt')
        coverage_wf(input_fp, map_fp, output_fp, threshold=80,
                    names_fp=names_fp)
        with open(output_fp, 'r') as f:
            obs = f.read().splitlines()
        self.assertEqual(len(obs), 86)
        for line in obs:
            if line.startswith('ARABCAT-PWY\t'):
                self.assertListEqual(line.split('\t')[1:], [
                    '0', '1', '1', '1', '1', 'L-arabinose degradation I'])
        remove(output_fp)

        # wrong mapping file
        map_fp = join(self.datdir, 'tree.nwk')
        with self.assertRaises(SystemExit) as ctx:
            coverage_wf(input_fp, map_fp, output_fp, count=True)
        errmsg = 'No group membership is found in tree.nwk.'
        self.assertEqual(str(ctx.exception), errmsg)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from unittest import TestCase, main
from os import remove
from os.path import join, dirname, realpath
from shutil import rmtree
from tempfile import mkdtemp
from io import StringIO

import numpy as np
import pandas as pd
from biom import load_table, Table
from pandas.testing import assert_frame_equal

from woltka.biom import (
    table_to_biom, biom_to_table, write_biom, biom_max_f, divide_biom,
    scale_biom, filter_biom, round_biom, biom_add_metacol, collapse_biom)
from woltka.table import prep_table


class BiomTests(TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()
        self.datdir = join(dirname(realpath(__file__)), 'data')

    def tearDown(self):
        rmtree(self.tmpdir)

    def test_table_to_biom(self):
        data = [[4, 2, 0],
                [5, 0, 3],
                [8, 0, 0],
                [0, 3, 0],
                [0, 7, 5]]
        observs = ['G1', 'G2', 'G3', 'G4', 'G5']
        samples = ['S1', 'S2', 'S3']
        metadata = [
            {'Name': 'Actinobacteria', 'Rank': 'phylum', 'Lineage': '2;72;74'},
            {'Name': 'Firmicutes',     'Rank': 'phylum', 'Lineage': '2;72'},
            {'Name': 'Bacteroidetes',  'Rank': 'phylum', 'Lineage': '2;70'},
            {'Name': 'Cyanobacteria',  'Rank': 'phylum', 'Lineage': '2;72'},
            {'Name': '',               'Rank': '',       'Lineage': ''}]
        obs = table_to_biom(data, observs, samples, metadata)
        exp = pd.read_csv(StringIO(
            '\tS1\tS2\tS3\tName\tRank\tLineage\n'
            'G1\t4\t2\t0\tActinobacteria\tphylum\t2;72;74\n'
            'G2\t5\t0\t3\tFirmicutes\tphylum\t2;72\n'
            'G3\t8\t0\t0\tBacteroidetes\tphylum\t2;70\n'
            'G4\t0\t3\t0\tCyanobacteria\tphylum\t2;72\n'
            'G5\t0\t7\t5\t\t\t\n'),
            sep='\t', index_col=0, na_filter=False)
        assert_frame_equal(
            obs.to_dataframe(dense=True).astype(int), exp.iloc[:, :3])
        assert_frame_equal(
            obs.metadata_to_dataframe('observation')[[
                'Name', 'Rank', 'Lineage']], exp.iloc[:, -3:])

    def test_biom_to_table(self):
        data = [[4, 2, 0],
                [5, 0, 3],
                [8, 0, 0],
                [0, 3, 0],
                [0, 7, 5]]
        observs = ['G1', 'G2', 'G3', 'G4', 'G5']
        samples = ['S1', 'S2', 'S3']
        metadata = [
            {'Name': 'Actinobacteria', 'Rank': 'phylum', 'Lineage': '2;72;74'},
            {'Name': 'Firmicutes',     'Rank': 'phylum', 'Lineage': '2;72'},
            {'Name': 'Bacteroidetes',  'Rank': 'phylum', 'Lineage': '2;70'},
            {'Name': 'Cyanobacteria',  'Rank': 'phylum', 'Lineage': '2;72'},
            {'Name': '',               'Rank': '',       'Lineage': ''}]
        table = table_to_biom(data, observs, samples, metadata)
        obs = biom_to_table(table)
        self.assertListEqual(obs[0], data)
        self.assertListEqual(obs[1], observs)
        self.assertListEqual(obs[2], samples)
        self.assertListEqual(obs[3], metadata)

    def test_write_biom(self):
        profile = {'S1': {'G1': 4, 'G2': 5, 'G3': 8},
                   'S2': {'G1': 2, 'G4': 3, 'G5': 7},
                   'S3': {'G2': 3, 'G5': 5}}
        exp = table_to_biom(*prep_table(profile))
        fp = join(self.tmpdir, 'tmp.biom')
        write_biom(exp, fp)
        obs = load_table(fp)
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')
        remove(fp)

    def test_biom_max_f(self):
        table = Table(*map(np.array, prep_table({
            'S1': {'G1': 0, 'G2': 1, 'G3': 200},
            'S2': {'G1': 1.5, 'G2': 2.475, 'G3': 8.12782},
            'S3': {'G1': 1e-5, 'G2': 33.905, 'G3': 3.1415926}})))
        self.assertEqual(biom_max_f(table), 7)

    def test_divide_biom(self):
        obs = Table(*map(np.array, prep_table({
            'S1': {'G1': 20, 'G2': 36, 'G3': 4},
            'S2': {'G1': 15, 'G2': 24, 'G3': 8},
            'S3': {'G1': 10, 'G2': 18, 'G3': 0}})))
        sizes = {'G1': 5, 'G2': 6, 'G3': 2}
        divide_biom(obs, sizes)
        exp = Table(*map(np.array, prep_table({
            'S1': {'G1': 4, 'G2': 6, 'G3': 2},
            'S2': {'G1': 3, 'G2': 4, 'G3': 4},
            'S3': {'G1': 2, 'G2': 3, 'G3': 0}})))
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')
        del(sizes['G3'])
        with self.assertRaises(KeyError):
            divide_biom(obs, sizes)

    def test_scale_biom(self):
        obs = Table(*map(np.array, prep_table({
            'S1': {'G1': 4, 'G2': 7, 'G3': 0},
            'S2': {'G1': 2, 'G2': 3, 'G3': 1}})))
        scale_biom(obs, 3)
        exp = Table(*map(np.array, prep_table({
            'S1': {'G1': 12, 'G2': 21, 'G3': 0},
            'S2': {'G1':  6, 'G2':  9, 'G3': 3}})))
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')

    def test_filter_biom(self):
        table = Table(*map(np.array, prep_table({
            'S1': {'G1': 4, 'G2': 5, 'G3': 8},
            'S2': {'G1': 2, 'G4': 3, 'G5': 7},
            'S3': {'G2': 3, 'G5': 5}})))
        obs = filter_biom(table, th=3)
        exp = Table(*map(np.array, prep_table({
            'S1': {'G1': 4, 'G2': 5, 'G3': 8},
            'S2': {'G4': 3, 'G5': 7},
            'S3': {'G2': 3, 'G5': 5}})))
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')

        obs = filter_biom(table, th=4)
        exp = Table(*map(np.array, prep_table({
            'S1': {'G1': 4, 'G2': 5, 'G3': 8},
            'S2': {'G5': 7},
            'S3': {'G5': 5}})))
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')

        obs = filter_biom(table, th=6)
        exp = Table(*map(np.array, prep_table({
            'S1': {'G3': 8},
            'S2': {'G5': 7},
            'S3': {}})))
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')

        obs = filter_biom(table, th=0.25)
        exp = Table(*map(np.array, prep_table({
            'S1': {'G2': 5, 'G3': 8},
            'S2': {'G4': 3, 'G5': 7},
            'S3': {'G2': 3, 'G5': 5}})))
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')

        obs = filter_biom(table, th=0.5)
        exp = Table(*map(np.array, prep_table({
            'S1': {},
            'S2': {'G5': 7},
            'S3': {'G5': 5}})))
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')

        # empty BIOM table cannot be directly compared
        obs = filter_biom(table, th=10)
        self.assertTupleEqual(obs.to_dataframe(True).shape, (0, 3))

    def test_round_biom(self):
        obs = Table(*map(np.array, prep_table({
            'S1': {'G1': 0.5, 'G2': 0.0, 'G3': 2.3, 'G4': 0.50000000001},
            'S2': {'G1': 1.5, 'G2': 0.2, 'G3': 1.49999999999, 'G4': 0.2},
            'S3': {'G1': 2.5, 'G2': 0.3, 'G3': 3.8, 'G4': 0.1}})))
        round_biom(obs)
        exp = Table(*map(np.array, prep_table({
            'S1': {'G1': 0, 'G3': 2},
            'S2': {'G1': 2, 'G3': 2},
            'S3': {'G1': 2, 'G3': 4}})))
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')
        obs = Table(*map(np.array, prep_table({
            'S1': {'G1': 0.225, 'G2': 0.0,   'G3': 2.375},
            'S2': {'G1': 1.547, 'G2': 0.173, 'G3': 1.499}})))
        round_biom(obs, 2)
        exp = Table(*map(np.array, prep_table({
            'S1': {'G1': 0.23, 'G2': 0.0,  'G3': 2.38},
            'S2': {'G1': 1.55, 'G2': 0.17, 'G3': 1.5}})))
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')

    def test_biom_add_metacol(self):
        obs = Table(*map(np.array, prep_table({
            'S1': {'G1': 4, 'G2': 5, 'G3': 8, 'G4': 0, 'G5': 3},
            'S2': {'G1': 1, 'G2': 8, 'G3': 0, 'G4': 7, 'G5': 4},
            'S3': {'G1': 0, 'G2': 2, 'G3': 3, 'G4': 5, 'G5': 0}})))
        self.assertIsNone(obs.metadata(axis='observation'))
        rankdic = {'G1': 'S', 'G2': 'S', 'G3': 'F', 'G4': 'O', 'G5': 'P'}
        biom_add_metacol(obs, rankdic, 'Rank')
        exp = [{'Rank': 'S'}, {'Rank': 'S'}, {'Rank': 'F'}, {'Rank': 'O'},
               {'Rank': 'P'}]
        self.assertListEqual(list(map(
            dict, obs.metadata(axis='observation'))), exp)
        namedic = {'G1': 'Proteo', 'G3': 'Actino', 'G2': 'Firmic',
                   'G4': 'Bacter'}
        biom_add_metacol(obs, namedic, 'Name', missing='X')
        exp = [{'Rank': 'S', 'Name': 'Proteo'},
               {'Rank': 'S', 'Name': 'Firmic'},
               {'Rank': 'F', 'Name': 'Actino'},
               {'Rank': 'O', 'Name': 'Bacter'},
               {'Rank': 'P', 'Name': 'X'}]
        self.assertListEqual(list(map(
            dict, obs.metadata(axis='observation'))), exp)

    def test_collapse_biom(self):
        table = Table(*map(np.array, prep_table({
            'S1': {'G1': 4, 'G2': 5, 'G3': 8, 'G4': 0, 'G5': 3, 'G6': 0},
            'S2': {'G1': 1, 'G2': 8, 'G3': 0, 'G4': 7, 'G5': 4, 'G6': 2},
            'S3': {'G1': 0, 'G2': 2, 'G3': 3, 'G4': 5, 'G5': 0, 'G6': 9}})))

        # one-to-one mapping (e.g., direct translation)
        mapping = {'G1': ['H1'], 'G2': ['H2'], 'G3': ['H3'],
                   'G4': ['H4'], 'G5': ['H5'], 'G6': ['H6']}
        obs = collapse_biom(table.copy(), mapping)
        exp = Table(*map(np.array, prep_table({
            'S1': {'H1': 4, 'H2': 5, 'H3': 8, 'H4': 0, 'H5': 3, 'H6': 0},
            'S2': {'H1': 1, 'H2': 8, 'H3': 0, 'H4': 7, 'H5': 4, 'H6': 2},
            'S3': {'H1': 0, 'H2': 2, 'H3': 3, 'H4': 5, 'H5': 0, 'H6': 9}})))
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')

        # some missing, some extra
        mapping = {'G1': ['H1'], 'G2': ['H2'], 'G3': ['H3'], 'G9': ['H9']}
        obs = collapse_biom(table.copy(), mapping)
        exp = Table(*map(np.array, prep_table({
            'S1': {'H1': 4, 'H2': 5, 'H3': 8},
            'S2': {'H1': 1, 'H2': 8, 'H3': 0},
            'S3': {'H1': 0, 'H2': 2, 'H3': 3}})))
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')

        # wrong mapping (no match)
        mapping = {'H1': ['I1'], 'H2': ['I2'], 'H3': ['I3']}
        obs = collapse_biom(table.copy(), mapping)
        self.assertTrue(obs.is_empty())
        self.assertListEqual(list(obs.ids('sample')), ['S1', 'S2', 'S3'])
        self.assertListEqual(list(obs.ids('observation')), [])

        # many-to-one mapping (e.g., taxonomic rank up)
        mapping = {'G1': ['H1'], 'G2': ['H1'], 'G3': ['H2'],
                   'G4': ['H2'], 'G5': ['H2'], 'G6': ['H3']}
        obs = collapse_biom(table.copy(), mapping)
        exp = Table(*map(np.array, prep_table({
            'S1': {'H1': 9, 'H2': 11, 'H3': 0},
            'S2': {'H1': 9, 'H2': 11, 'H3': 2},
            'S3': {'H1': 2, 'H2':  8, 'H3': 9}})))
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')

        # many-to-many mapping (e.g., genes to pathways)
        mapping = {'G1': ['H1'],
                   'G2': ['H1', 'H2'],
                   'G3': ['H2', 'H3', 'H4'],
                   'G4': ['H2', 'H5'],
                   'G5': ['H4'],
                   'G6': ['H3', 'H5']}
        obs = collapse_biom(table.copy(), mapping)
        exp = Table(*map(np.array, prep_table({
            'S1': {'H1': 9, 'H2': 13, 'H3':  8, 'H4': 11, 'H5':  0},
            'S2': {'H1': 9, 'H2': 15, 'H3':  2, 'H4':  4, 'H5':  9},
            'S3': {'H1': 2, 'H2': 10, 'H3': 12, 'H4':  3, 'H5': 14}})))
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')

        # many-to-many mapping, with normalization
        obs = collapse_biom(table.copy(), mapping, divide=True)
        exp = Table(*map(np.array, prep_table({
            'S1': {'H1': 6, 'H2': 5, 'H3': 3, 'H4': 6, 'H5': 0},
            'S2': {'H1': 5, 'H2': 8, 'H3': 1, 'H4': 4, 'H5': 4},
            'S3': {'H1': 1, 'H2': 4, 'H3': 6, 'H4': 1, 'H5': 7}})))
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')

        # nothing left after normalization
        table = Table(*map(np.array, prep_table({
            'S1': {'G1': 0}, 'S2': {'G1': 1}, 'S3': {'G1': 2}})))
        mapping = {'G1': ['H1', 'H2', 'H3', 'H4']}
        obs = collapse_biom(table.copy(), mapping, divide=True)
        self.assertTrue(obs.is_empty())
        self.assertListEqual(list(obs.ids('sample')), ['S1', 'S2', 'S3'])
        self.assertListEqual(list(obs.ids('observation')), [])

        # stratified table
        table = Table(*map(np.array, prep_table({
            'S1': {'A|K1': 4, 'A|K2': 5, 'B|K2': 8, 'C|K3': 3, 'C|K4': 0},
            'S2': {'A|K1': 1, 'A|K2': 8, 'B|K2': 0, 'C|K3': 4, 'C|K4': 2}})))
        mapping = {'K1': ['H1'], 'K2': ['H2', 'H3'], 'K3': ['H3']}
        obs = collapse_biom(table.copy(), mapping, field=1)
        exp = Table(*map(np.array, prep_table({
            'S1': {'A|H1': 4, 'A|H2': 5, 'A|H3': 5, 'B|H2': 8, 'B|H3': 8,
                   'C|H3': 3},
            'S2': {'A|H1': 1, 'A|H2': 8, 'A|H3': 8, 'B|H2': 0, 'B|H3': 0,
                   'C|H3': 4}})))
        self.assertEqual(obs.descriptive_equality(exp), 'Tables appear equal')

        # invalid field
        with self.assertRaises(ValueError) as ctx:
            collapse_biom(table, mapping, field=2)
        errmsg = 'Feature "A|K1" has less than 3 fields.'
        self.assertEqual(str(ctx.exception), errmsg)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

"""Main classification workflow.

Notes
-----
Only in this script can functions directly interface with the user by screen
output (via `click`) and file input/output, except for raising errors.
"""

from os import makedirs
from os.path import join, basename, isfile, isdir
from collections import deque, defaultdict
from functools import partial, lru_cache
from typing import Tuple
import click

from .util import (
    update_dict, allkeys, sum_dict, scale_factor, scale_dict, round_dict)
from .file import (
    openzip, readzip, path2stem, stem2rank, read_ids, id2file_from_dir,
    id2file_from_map, read_map_uniq, read_map_1st, write_readmap)
from .align import plain_mapper, range_mapper
from .classify import (
    assign_none, assign_free, assign_rank, counter, counter_size,
    counter_strat, counter_size_strat)
from .tree import (
    read_names, read_nodes, read_lineage, read_newick, read_columns,
    fill_root)
from .ordinal import (
    ordinal_mapper, load_gene_coords, calc_gene_lens)
from .table import prep_table, write_table
from .coverage import parse_ranges, calc_coverage, write_coverage


def workflow(input_fp:     str,
             output_fp:    str,
             # input
             input_fmt:    str = None,
             input_ext:    str = None,
             samples:      str = None,
             demux:       bool = None,
             trimsub:      str = None,
             # hierarchies
             nodes_fps:   list = [],
             newick_fps:  list = [],
             lineage_fps: list = [],
             columns_fps: list = [],
             map_fps:     list = [],
             map_rank:    bool = False,
             names_fps:   list = [],
             # assignment
             ranks:        str = None,
             uniq:        bool = False,
             major:       bool = None,
             above:       bool = False,
             subok:       bool = False,
             # gene matching
             coords_fp:    str = None,
             overlap:      int = 80,
             # stratification
             strata_dir:   str = None,
             # normalization
             sizes:        str = None,
             frac:        bool = False,
             scale:        str = None,
             digits:       int = None,
             # output
             output_fmt:   str = None,
             unassigned:  bool = False,
             name_as_id:  bool = False,
             add_rank:    bool = False,
             add_lineage: bool = False,
             outmap_dir:   str = None,
             outmap_zip:   str = 'gz',
             outcov_dir:   str = None,
             # performance
             chunk:        int = None,
             cache:        int = 1024,
             no_exe:      bool = False) -> dict:
    """Main classification workflow which accepts command-line arguments.

    Returns
    -------
    dict
        Resulting profile.

    Notes
    -----
    This function directly parses groups of command-line parameters to the
    following sub-workflows. Paramters are briefly explained in `cli.classify`
    and detailed in individual sub-workflows.

    See Also
    --------
    .cli.classify
        Command-line arguments and help information.
    """
    # available external compressors
    zippers = None if no_exe else {}

    # parse input samples
    samples, files, demux = parse_samples(
        input_fp, input_ext, samples, demux)

    # parse stratification files
    stratmap = parse_strata(strata_dir, samples)

    # build classification system
    tree, rankdic, namedic, root = build_hierarchy(
        names_fps, nodes_fps, newick_fps, lineage_fps, columns_fps, map_fps,
        map_rank, zippers)

    # build mapping module
    mapper, chunk = build_mapper(
        coords_fp, outcov_dir, overlap, chunk, zippers)

    # parse length map
    sizes = parse_sizes(sizes, mapper, zippers)

    # target classification ranks
    ranks, rank2dir = prepare_ranks(ranks, outmap_dir, tree, rankdic)

    # classify query sequences
    data = classify(
        mapper, files, samples, input_fmt, demux, trimsub, tree, rankdic,
        namedic if name_as_id else None, root, ranks, rank2dir, outmap_zip,
        uniq, major, above, subok, sizes, unassigned, stratmap, chunk, cache,
        zippers, outcov_dir)

    # convert counts to fractions
    frac_profiles(data, frac)

    # scale values
    scale_profiles(data, scale)

    # round values and drop zeros
    round_profiles(data, digits)

    # write output profiles
    write_profiles(
        data, output_fp, output_fmt, samples, tree, rankdic, namedic,
        name_as_id, add_rank, add_lineage)

    click.echo('Task completed.')
    return data


def classify(mapper:  object,
             files:     list or dict,
             samples:   list = None,
             fmt:        str = None,
             demux:     bool = None,
             trimsub:    str = None,
             tree:      dict = None,
             rankdic:   dict = None,
             namedic:   dict = None,
             root:       str = None,
             ranks:      str = None,
             rank2dir:  dict = None,
             outzip:     str = None,
             uniq:      bool = False,
             major:      int = None,
             above:     bool = False,
             subok:     bool = False,
             sizes:     dict = None,
             unasgd:    bool = False,
             stratmap:  dict = None,
             chunk:      int = None,
             cache:      int = 1024,
             zippers:   dict = None,
             outcov_dir: str = None) -> dict:
    """Core of the classification workflow.

    Parameters
    ----------
    mapper : object
        Mapping module (Plain or Ordinal).
    files : list or dict
        Paths to input alignment files, if multiplexed, or dictionary of file
        paths to sample IDs, if per-sample.
    samples : list of str, optional
        Sample ID list to include.
    fmt : str, optional
        Format of input alignment file. Options:
        - 'b6o': BLAST tabular format.
        - 'sam': SAM format.
        - 'map': Simple map of query <tab> subject.
        If None, program will automatically infer from file content.
    demux : bool, optional
        Whether perform demultiplexing.
    trimsub : str, optional
        Trim subject IDs at the last given delimiter.
    tree : dict, optional
        Taxonomic tree.
    rankdic : dict, optional
        Rank dictionary.
    namedic : dict, optional
        Taxon name dictionary.
    root : str, optional
        Root identifier.
    ranks: list of str, optional
        List of ranks at each of which sequences are to be classified. Can also
        be "none" to omit classification (simply report subject IDs) or "free"
        to perform free-rank classification (LCA of subjects regardless of rank
        will be reported).
    rank2dir : dict, otional
        Write classification map per rank to directory.
    outzip : str, optional
        Output read map compression method (gz, bz2, xz or None).
    uniq : bool, optional
        Assignment must be unique. Otherwise, report all possible assignments
        and normalize counts (for none- and fixed-rank assignments).
    major : int, optional
        In given-rank classification, perform majority-rule assignment based on
        this percentage threshold. Range: [51, 99].
    above : bool, optional
        Allow assigning to a classification unit higher than given rank.
    subok : bool, optional
        In free-rank classification, allow assigning sequences to their direct
        subjects instead of higher classification units, if applicable.
    sizes : dict, optional
        Subject size dictionary.
    unasgd : bool, optional
        Report unassigned sequences.
    stratmap : dict, optional
        Map of sample ID to stratification file.
    chunk : int, optional
        Number of lines per chunk to read from alignment file.
    cache : int, optional
        LRU cache size for classification results at each rank.
    zippers : dict, optional
        External compression programs.
    outcov_dir : str, optional
        Write Subject coverage maps to directory.

    Returns
    -------
    dict of dict
        Per-rank profiles generated from classification.

    Notes
    -----
    Subject(s) of each query are sorted and converted into a tuple, which is
    hashable, a property necessary for subsequent assignment result caching.
    """

    data = {x: {} for x in ranks}

    # assigners for each rank
    assigners = {}

    # assignment parameters
    kwargs = {'assigners': assigners, 'cache': cache, 'tree': tree, 'rankdic':
              rankdic, 'namedic': namedic, 'root':  root, 'uniq': uniq,
              'major': major and major / 100, 'above': above, 'subok': subok,
              'sizes': sizes, 'unasgd': unasgd, 'rank2dir': rank2dir,
              'outzip': outzip if outzip != 'none' else None}

    # (optional) subject coverage data
    covers = {} if outcov_dir else None

    # current sample Id
    csample = False

    # parse input alignment file(s) and generate profile(s)
    for fp in sorted(files):

        # decide input type (stdin or file)
        if fp == '-':
            fileobj = click.open_file(fp).__iter__()
            click.echo('Parsing alignment from stdin ', nl=False)
        else:
            fileobj = readzip(fp, zippers)
            click.echo(f'Parsing alignment file {basename(fp)} ', nl=False)

        # read alignment file into query-to-subject(s) map
        with fileobj as fh:

            # query and progress counters
            nqry, nstep = 0, -1

            # parse alignment file by chunk
            for qryque, subque in mapper(fh, fmt=fmt, n=chunk):
                nqry += len(qryque)

                # (optional) demultiplex and generate per-sample maps
                rmaps = demultiplex(qryque, subque, samples) if demux else {
                    files[fp] if files else None: (qryque, subque)}

                # (optional) calculate subject coverage
                if outcov_dir:
                    parse_ranges(rmaps, covers)

                # assign reads at each rank
                for sample, (qryque, subque) in rmaps.items():

                    # (optional) strip suffixes from subject Ids
                    subque = deque(map(tuple, map(sorted, strip_suffix(
                        subque, trimsub) if trimsub else subque)))

                    # (optional) read strata of current sample into cache
                    if stratmap and sample != csample:
                        kwargs['strata'] = read_strata(
                            stratmap[sample], zippers)
                        csample = sample

                    # call assignment workflow for each rank
                    for rank in ranks:
                        assign_readmap(
                            qryque, subque, data, rank, sample, **kwargs)

                # show progress
                istep = nqry // 1000000 - nstep
                if istep:
                    click.echo('.' * istep, nl=False)
                    nstep += istep

        click.echo(' Done.')
        click.echo(f'  Number of sequences classified: {nqry}.')

    # write coverage maps
    if outcov_dir:
        click.echo('Calculating per sample coverage...', nl=False)
        write_coverage(calc_coverage(covers), outcov_dir)
        click.echo(' Done.')

    click.echo('Classification completed.')
    return data


def parse_samples(fp:        str,
                  ext:       str = None,
                  samples:   str = None,
                  demux:    bool = None) -> Tuple[list, list or dict, bool]:
    """Determine sample IDs, aligment files, and multiplex status.

    Parameters
    ----------
    fp : str
        Path to a file or a directory.
    ext : str, optional
        Filename extension.
    samples : str, optional
        Comma-separated list of sample IDs, or path to sample ID list file.
    demux : bool, optional
        Whether perform demultiplexing.

    Returns
    -------
    list
        Sample IDs to include.
    list or dict
        Filepaths if demultiplexing, or filepath to sample ID map if not.
    bool
        Whether perform demultiplexing.
    """
    # read sample Ids
    if samples:
        samples = read_ids(samples) if isfile(samples) else samples.split(',')
        click.echo(f'Number of samples to include: {len(samples)}.')

    errmsg = 'Provided sample IDs and actual files are inconsistent.'

    # input is stdin
    if fp == '-':

        # turn on demultiplexing if not decided
        demux = demux is not False
        if demux:
            files = [fp]

        # sample Id is empty
        else:
            files = {fp: ''}
            samples = ['']

        click.echo('Input alignment is from stdin.')

    # path is a directory
    elif isdir(fp):

        # turn off demultiplexing if not decided
        demux = demux or False

        # get a map of plausible sample Ids to files
        map_ = id2file_from_dir(fp, ext, not demux and samples)
        if len(map_) == 0:
            raise ValueError('No valid file found in directory.')
        if demux:
            files = sorted([join(fp, x) for x in map_.values()])

        # validate with given sample Ids
        else:
            if not samples:
                samples = sorted(map_.keys())
            elif len(map_) < len(samples):
                raise ValueError(errmsg)
            files = {join(fp, map_[x]): x for x in samples}

        click.echo(f'Input directory: {fp}.')
        click.echo(f'Number of alignment files to read: {len(files)}.')

    # path is a file
    elif isfile(fp):

        # check if file is an Id-to-file map
        map_ = id2file_from_map(fp)
        if map_:

            # turn off demultiplexing if not decided
            demux = demux or False

            # validate with given sample Ids
            if samples:
                map_ = dict(map_)
                try:
                    files = {map_[x]: x for x in samples}
                except KeyError:
                    raise ValueError(errmsg)

            # sample order and files provided
            else:
                samples = [x[0] for x in map_]
                files = {x[1]: x[0] for x in map_}

            click.echo(f'Number of alignment files to read: {len(files)}.')

        # treat file as a single alignment file
        else:

            # turn on demultiplexing if not decided
            demux = demux is not False
            if demux:
                files = [fp]

            # validate with given sample Ids
            else:
                sample = path2stem(fp, ext)
                if samples and samples != [sample]:
                    raise ValueError(errmsg)
                files = {fp: sample}
                samples = [sample]

            click.echo(f'Input alignment file: {fp}.')

    else:
        raise ValueError(f'"{fp}" is not a valid file or directory.')

    click.echo(f'Demultiplexing: {"on" if demux else "off"}.')

    return samples, files, demux


def parse_strata(fp:       str = None,
                 samples: list = None) -> dict:
    """Get a map of sample Ids to mapping files for stratification.

    Parameters
    ----------
    fp : str, optional
        Directory of read-to-feature maps for stratification.

    Returns
    -------
    dict or None
        Sample ID to stratification file map, or None if not applicable.

    Raises
    ------
    ValueError
        Sample IDs are given but stratification files are missing for one or
        more samples.
    """
    if not fp:
        return
    click.echo(f'Stratification file directory: {fp}.')
    map_ = id2file_from_dir(fp, ids=samples)
    if len(samples or []) > len(map_):
        raise ValueError(
            'Cannot locate stratification files for one or more samples.')
    return {k: join(fp, v) for k, v in map_.items()}


def build_mapper(coords_fp:  str = None,
                 outcov_dir: str = None,
                 overlap:    int = None,
                 chunk:      int = None,
                 zippers:   dict = None) -> Tuple[callable, int]:
    """Build mapper function (plain or ordinal).

    Parameters
    ----------
    coords_fp : str, optional
        Path to gene coordinates file.
    outcov_dir : str, optional
        Path to output subject coverage directory.
    overlap : int, optional
        Read/gene overlapping percentage threshold.
    chunk : int, optional
        Number of lines per chunk to read from alignment file.
    zippers : dict, optional
        External compression programs.

    Returns
    -------
    callable
        Mapper function.
    int
        Number of lines per chunk.

    Notes
    -----
    Currently two mappers are supported: "plain" for regular alignments
    (i.e., simple query-to-subject maps), "ordinal" for alignments with
    coordinates which will be used to match queries (reads) and genes. The
    presence of a gene coordinates file (`coords_fp`) is an indicator for
    using the latter.

    The chunk size, if not specified, is determined empirically: 1,000 for
    plain mapper and 1,000,000 for ordinal mapper.
    """
    if coords_fp:
        click.echo('Reading gene coordinates...', nl=False)
        with readzip(coords_fp, zippers) as fh:
            coords, idmap, prefix = load_gene_coords(fh, sort=True)
        click.echo(' Done.')
        click.echo(f'  Total number of host sequences: {len(coords)}.')
        chunk = chunk or 1000000
        return partial(ordinal_mapper, coords=coords, idmap=idmap,
                       prefix=prefix, th=overlap and overlap / 100), chunk
    else:
        chunk = chunk or 1000
        return range_mapper if outcov_dir else plain_mapper, chunk


def parse_sizes(sizes:       str,
                mapper: callable,
                zippers:    dict = None) -> dict or None:
    """Generate a feature to length mapping.

    Parameters
    ----------
    sizes : str
        Path to feature size mapping file.
    mapper : callable
        Read/gene overlapping percentage threshold.
    zippers : dict, optional
        External compression programs.

    Returns
    -------
    dict or None
        Feature to length mapping.

    Raises
    ------
    ValueError
        Sizes is set as "." but gene coordinates file is not provided.
    """
    if not sizes:
        return

    # calculate gene lengths by coordinates
    if sizes == '.':
        click.echo('Calculating gene lengths from coordinates...', nl=False)
        try:
            sizemap = calc_gene_lens(mapper)
        except AttributeError:
            raise ValueError('Gene coordinates file is not provided.')
        click.echo(' Done.')

    # read sizes from file
    else:
        click.echo(f'Reading subject sizes file: {basename(sizes)}...',
                   nl=False)
        with readzip(sizes, zippers) as f:
            sizemap = {k: float(v) for k, v in read_map_1st(f)}
        click.echo(' Done.')

    # convert to reciprocal
    return {k: 1 / v for k, v in sizemap.items()}


def prepare_ranks(ranks:      str = None,
                  outmap_dir: str = None,
                  tree:      dict = None,
                  rankdic:   dict = None) -> Tuple[list, dict or None]:
    """Prepare classification ranks and directories of read-to-feature maps.

    Parameters
    ----------
    ranks : str, optional
        Target ranks (comma-separated).
    outmap_dir : str, optional
        Path to output read map directory.
    tree : dict, optional
        Taxonomic tree.
    rankdic : dict, optional
        Rank dictionary.

    Returns
    -------
    list
        Ranks.
    dict or None
        Rank-to-directory map (or None if not necessary).
    """
    # ranks are given
    if ranks:
        ranks = ranks.split(',')

        # check presence of ranks in classification system
        if rankdic is not None:
            missing = set(ranks) - {'free', 'none'} - set(rankdic.values())
            if missing:
                raise ValueError(f'Ranks {", ".join(sorted(missing))} are not'
                                 ' found in classification system.')

    # if classification system is provided, do free-rank classification;
    # otherwise do no-rank assignment.
    else:
        ranks = ['free' if tree else 'none']

    click.echo('Classification will operate on these ranks: {}.'.format(
        ', '.join(ranks)))

    # check output directory
    if not outmap_dir:
        return ranks, None
    makedirs(outmap_dir, exist_ok=True)
    click.echo(f'Read-to-feature maps will be saved to: {outmap_dir}.')

    # determine output read map directory per rank
    if len(ranks) == 1:
        rank2dir = {ranks[0]: outmap_dir}
    else:
        rank2dir = {}
        for rank in ranks:
            dir_ = join(outmap_dir, rank)
            makedirs(dir_, exist_ok=True)
            rank2dir[rank] = dir_

    return ranks, rank2dir


def build_hierarchy(names_fps:   list = [],
                    nodes_fps:   list = [],
                    newick_fps:  list = [],
                    lineage_fps: list = [],
                    columns_fps: list = [],
                    map_fps:     list = [],
                    map_rank:    bool = None,
                    zippers:     dict = None) -> Tuple[dict, dict, dict, str]:
    """Construct hierarchical classification system.

    Parameters
    ----------
    names_fps : list of str, optional
        Taxonomic names file(s).
    nodes_fps : list of str, optional
        Taxonomic nodes file(s).
    newick_fps : list of str, optional
        Newick tree file.
    lineage_fps : list of str, optional
        Lineage strings file.
    columns_fps : list of str, optional
        Rank-per-column file.
    map_fps : list of str, optional
        Mapping file(s).
    map_rank : bool, optional
        Treat mapping filename stem as rank.
    zippers : dict, optional
        External compression programs.

    Returns
    -------
    dict
        Taxonomic tree.
    dict
        Rank dictionary.
    dict
        Name dictionary.
    str
        Root identifier.
    """
    tree, rankdic, namedic = {}, {}, {}

    # check if at least one filepath is specified
    is_build = any([
        names_fps, nodes_fps, newick_fps, lineage_fps, columns_fps, map_fps])
    if is_build:
        click.echo('Constructing classification system...')

    # taxonomy names
    for fp in names_fps:
        click.echo(f'  Parsing taxon names file: {basename(fp)}...', nl=False)
        with readzip(fp, zippers) as f:
            names = read_names(f)
        update_dict(namedic, names)
        click.echo(' Done.')

    # taxonomy nodes
    for fp in nodes_fps:
        click.echo(f'  Parsing taxon nodes file: {basename(fp)}...', nl=False)
        with readzip(fp, zippers) as f:
            tree_, rankdic_ = read_nodes(f)
        update_dict(tree, tree_)
        update_dict(rankdic, rankdic_)
        click.echo(' Done.')

    # Newick-format tree
    for fp in newick_fps:
        click.echo(f'  Parsing Newick tree file: {basename(fp)}...', nl=False)
        with readzip(fp, zippers) as f:
            update_dict(tree, read_newick(f))
        click.echo(' Done.')

    # lineage file
    for fp in lineage_fps:
        click.echo(f'  Parsing lineage file: {basename(fp)}...', nl=False)
        with readzip(fp, zippers) as f:
            tree_, rankdic_ = read_lineage(f)
        update_dict(tree, tree_)
        update_dict(rankdic, rankdic_)
        click.echo(' Done.')

    # columns file
    for fp in columns_fps:
        click.echo(f'  Parsing columns file: {basename(fp)}...', nl=False)
        with readzip(fp, zippers) as f:
            tree_, rankdic_ = read_columns(f)
            update_dict(tree, tree_)
            update_dict(rankdic, rankdic_)
        click.echo(' Done.')

    # whether to extract rank from filename
    if map_rank is None:
        map_rank = bool(map_fps) and not any([
            nodes_fps, newick_fps, lineage_fps, columns_fps])
    if map_rank:
        click.echo('  Will extract rank name from map filename.')

    # plain mapping files
    for fp in map_fps:
        click.echo(f'  Parsing simple map file: {basename(fp)}...', nl=False)
        with readzip(fp, zippers) as f:
            map_ = dict(read_map_1st(f))
        update_dict(tree, map_)

        # filename stem as rank
        if map_rank:
            rank = stem2rank(path2stem(fp))
            update_dict(rankdic, {k: rank for k in set(map_.values())})
        click.echo(' Done.')

    # fill root
    root = fill_root(tree)

    if is_build:
        click.echo('Classification system constructed.')
        click.echo(f'  Total number of classification units: {len(tree)}.')

    return tree, rankdic, namedic, root


def strip_suffix(subque: list,
                 sep:     str) -> object:
    """Strip suffixes from subject IDs at the last separator.

    Parameters
    ----------
    subque : iterable
        Subject(s) queue to manipulate.
    sep : str, optional
        Separator between subject ID and suffix.

    Returns
    -------
    generator of set
        Processed subject(s) queue.

    Notes
    -----
    This function will find the last occurrence of separator in a subject Id,
    and trim from it to the right end. If not found, the whole subject Id will
    be retained.
    """
    return map(set, map(partial(
        map, lambda x: x.rsplit(sep, 1)[0]), subque))


def demultiplex(qryque:  list,
                subque:  list,
                samples: list = None,
                sep:      str = '_') -> dict:
    """Demultiplex a query-to-subject(s) map.

    Parameters
    ----------
    qryque : iterable
        Query queue to demultiplex.
    subque : iterable
        Corresponding subject(s) queue.
    samples : iterable of str, optional
        Sample IDs to keep.
    sep : str, optional
        Separator between sample ID and read ID.

    Returns
    -------
    dict of (deque, deque)
        Per-sample read-to-subject(s) maps.

    Notes
    -----
    In a multiplexed alignment file, query IDs are composed of sample ID and
    read ID, separated by a character (default: "_"). This function separates
    them at the first occurrence of the separator from left. If the separator
    is not found, the entire query ID will be retained as read ID and sample
    ID will be `None`.
    """
    if samples:
        samset = set(samples)

    # per-sample read and subject(s) queues
    qryques, subques = defaultdict(deque), defaultdict(deque)

    # current sample Id (it can be None so start with False)
    csample = False

    # list append method references
    qry_add, sub_add = None, None

    for query, subjects in zip(qryque, subque):

        # split query Id by first separator
        left, _, right = query.partition(sep)

        # if separator is present, take left and right
        # if there is no separator, take None and left
        sample, read = right and left, right or left

        # append read Id and subject(s) to queues
        if sample == csample:
            qry_add(read)
            sub_add(subjects)

        # (re-)assign method references to current sample
        elif not samples or sample in samset:
            csample = sample
            qry_add = qryques[sample].append
            sub_add = subques[sample].append

            qry_add(read)
            sub_add(subjects)

    return {x: (qryques[x], subques[x]) for x in qryques}


def read_strata(strata_fp: str,
                zippers:  dict = None) -> dict:
    """Read a stratification file for a sample.

    Parameters
    ----------
    strata_fp : str,
        Path to stratification file.
    zippers : dict, optional
        External compression programs.

    Returns
    -------
    dict
        Stratification information.

    Raises
    ------
    ValueError
        No stratification information is found in file.
    """
    with readzip(strata_fp, zippers) as fhs:
        strata = dict(read_map_uniq(fhs))
    if not strata:
        raise ValueError('No stratification information is found in file: '
                         f'{basename(strata_fp)}.')
    return strata


def assign_readmap(qryque:    list,
                   subque:    list,
                   data:      dict,
                   rank:       str,
                   sample:     str,
                   assigners: dict,
                   cache:      int = 1024,
                   rank2dir:  dict = None,
                   outzip:     str = None,
                   tree:      dict = None,
                   rankdic:   dict = None,
                   namedic:   dict = None,
                   root:       str = None,
                   uniq:      bool = False,
                   major:    float = None,
                   above:     bool = False,
                   subok:     bool = False,
                   sizes:     dict = None,
                   unasgd:    bool = False,
                   strata:    dict = None):
    """Assign query sequences in a query-to-subjects map to classification
    units based on their subjects.

    Parameters
    ----------
    qryque : iterable of str
        Query queue to assign.
    subque : iterable of tuple
        Subject(s) queue for assignment.
    data : dict
        Master data structure.
    rank : str
        Target rank to assign to.
    sample : str
        Sample ID.
    assigners : dict of callable
        Per-rank assigners.
    cache : int, optional
        LRU cache size for classification results at each rank.
    rank2dir : dict, optional
        Directory of output maps per rank.
    outzip : str, optional
        Output read map compression method (gz, bz2, xz or None).
    tree : dict, optional
        Hierarchical classification system.
    rankdic : dict, optional
        Rank dictionary.
    namedic : dict
        Taxon name directory.
    root : str, optional
        Root identifier.
    uniq : bool, optional
        Assignment must be unique. Otherwise, report all possible assignments
        and normalize counts (for none- and fixed-rank assignments).
    major : float, optional
        In given-rank classification, perform majority-rule assignment based on
        this fraction threshold.
    above : bool, optional
        In given-rank classification, assignment above the specified rank is
        acceptable.
    subok : bool, optional
        In free-rank classification, allow assigning sequences to their direct
        subjects instead of higher classification units, if applicable.
    sizes : dict, optional
        Subject size dictionary, with which counts will be divided by sizes.
    unasgd : bool, optional
        Report unassigned sequences.
    strata : dict, optional
        Read-to-feature map for stratification.

    Raises
    ------
    ValueError
        One or more subjects are not found in the size map.
    """
    # determine assigner and initiate (if not already)
    if rank is None or rank == 'none' or tree is None:
        if 'none' not in assigners:
            assigners['none'] = lru_cache(maxsize=cache)(partial(
                assign_none, uniq=uniq))
        assigner = assigners['none']
    elif rank == 'free':
        if 'free' not in assigners:
            assigners['free'] = lru_cache(maxsize=cache)(partial(
                assign_free, tree=tree, root=root, subok=subok))
        assigner = assigners['free']
    else:
        if rank not in assigners:
            assigners[rank] = lru_cache(maxsize=cache)(partial(
                assign_rank, rank=rank, tree=tree, rankdic=rankdic, root=root,
                major=major, above=above, uniq=uniq))
        assigner = assigners[rank]

    # call assigner on subject(s) per query
    taxque = map(assigner, subque)

    # report or drop unassigned
    if unasgd:
        taxque = (x or 'Unassigned' for x in taxque)

    # write classification map
    if rank2dir is not None:
        taxque = list(taxque)
        outfp = join(rank2dir[rank], f'{sample}.txt')
        with openzip(f'{outfp}.{outzip}' if outzip else outfp, 'at') as fh:
            write_readmap(fh, qryque, taxque, namedic)

    # determine counter and count taxon assignments
    try:
        counts = ((counter(taxque) if not sizes else
                   counter_size(subque, taxque, sizes)) if not strata else
                  (counter_strat(qryque, taxque, strata) if not sizes else
                   counter_size_strat(qryque, subque, taxque, sizes, strata)))
    except KeyError:
        raise ValueError('One or more subjects are not found in the size map.')

    # combine old and new counts
    sum_dict(data[rank].setdefault(sample, {}), counts)


def frac_profiles(data: dict,
                  frac: bool = False):
    """Convert cell values to fractions (i.e., divide by sum of each sample).

    Parameters
    ----------
    data : dict
        Profile data.
    frac : bool, optional
        Whether convert counts to fractions.

    Notes
    -----
    Samples with only zeros will be skipped.
    """
    if not frac:
        return
    for profile in data.values():
        for sample in profile.values():
            total = sum(sample.values())
            if not total:
                continue
            for feature in sample:
                sample[feature] /= total


def scale_profiles(data: dict,
                   scale: str = None):
    """Scale cell values in profiles by a factor.

    Parameters
    ----------
    data : dict
        Profile data.
    scale : str, optional
        Scale factor.
    """
    if not scale:
        return
    scale = scale_factor(scale)
    for rank in data:
        for datum in data[rank].values():
            scale_dict(datum, scale)


def round_profiles(data:  dict,
                   digits: int = None):
    """Round cell values in profiles into integers, and drop zero values.

    Parameters
    ----------
    data : dict
        Profile data.
    digits : int, optional
        Digits after the decimal point.
    """
    for rank in data:
        for datum in data[rank].values():
            round_dict(datum, digits)


def write_profiles(data:        dict,
                   fp:           str,
                   is_biom:     bool = None,
                   samples:     list = None,
                   tree:        dict = None,
                   rankdic:     dict = None,
                   namedic:     dict = None,
                   name_as_id:  bool = False,
                   add_rank:    bool = False,
                   add_lineage: bool = False):
    """Write profile to an output table file.

    Parameters
    ----------
    data : dict
        Profile data.
    fp : str
        Path to output file or directory.
    is_biom : bool, optional
        Output BIOM or TSV format.
    samples : list, optional
        Ordered sample ID list.
    tree : dict, optional
        Taxonomic tree.
    rankdic : dict, optional
        Rank dictionary.
    namedic : dict, optional
        Taxon name dictionary.
    name_as_id : bool, optional
        Replace feature IDs with names.
    add_lineage: bool optional
        Append feature ranks to table.
    add_lineage: bool optional
        Append lineage strings to table.

    Notes
    -----
    The boolean parameter `is_biom` can be one of the three values:
    - `None` (default): To be auto-determined based on the user-supplied output
    filename extension. If ".biom", do BIOM, otherwise do TSV. If output path
    is a directory, do BIOM by default.
    - `True` (command-line flag `--to-biom`): BIOM format.
    - `False` (command-line flag `--to-tsv`): TSV format.
    """
    if not fp:
        return

    # determine sample order
    if not samples:
        samples = sorted(allkeys(data))

    # determine ranks
    ranks = sorted(data.keys())

    # determine output filename and format
    if len(ranks) == 1:
        # single output file
        rank2fp = {ranks[0]: fp}
        if is_biom is None:
            is_biom = fp.endswith('.biom')
    else:
        # multiple output files
        makedirs(fp, exist_ok=True)
        is_biom = is_biom is not False
        ext = 'biom' if is_biom else 'tsv'
        rank2fp = {x: join(fp, f'{x}.{ext}') for x in ranks}
    click.echo('Format of output feature table(s): {}.'.format(
        'BIOM' if is_biom else 'TSV'))

    # override Id-to-name conversion
    if namedic is None:
        name_as_id = False

    # write output profile(s)
    fmt = 'BIOM' if is_biom else 'TSV'
    click.echo(f'Writing output profiles in {fmt} format...')
    for rank, fp in rank2fp.items():
        table = prep_table(data[rank], samples, tree if add_lineage else None,
                           rankdic if add_rank else None, namedic, name_as_id)
        write_table(table, fp, is_biom)
        n, m = len(table[2]), len(table[1])
        click.echo(f'  Rank: {rank}, samples: {n}, features: {m}.')

    click.echo('Profiles written.')
